# -*- coding: utf-8 -*-
MODULE_DOCSTRING = """
SAC header specification, including documentation.

Header names, order, types, and nulls, as well as allowed enumerated values,
are specified here.  Header name strings, and their array order are contained
in separate float, int, and string tuples.  Enumerated values, and their
allowed string and integer values, are in dictionaries.  Header value
documentation is in a dictionary, for reuse throughout the package.

"""

# header documentation is large and used in several places, so we just write
# it once here and distributed it as needed.
DOC = {
    'npts': 'N    Number of points per data component.',
    'nvhdr': '''N    Header version number. Current value is the integer 6.
                  Older version data (NVHDR < 6) are automatically updated
                  when read into sac.''',
    'b': 'F    Beginning value of the independent variable.',
    'e': 'F    Ending value of the independent variable.',
    'iftype': '''I    Type of file:

                  * ITIME {Time series file}
                  * IRLIM {Spectral file---real and imaginary}
                  * IAMPH {Spectral file---amplitude and phase}
                  * IXY {General x versus y data}
                  * IXYZ {General XYZ (3-D) file}''',
    'leven': 'L    TRUE if data is evenly spaced.',
    'delta': 'F    Increment between evenly spaced samples (nominal value).',
    'odelta': 'F    Observed increment if different from nominal value.',
    'idep': '''I    Type of dependent variable:

                  * IUNKN (Unknown)
                  * IDISP (Displacement in nm)
                  * IVEL (Velocity in nm/sec)
                  * IVOLTS (Velocity in volts)
                  * IACC (Acceleration in nm/sec/sec)''',
    'scale': 'F    Multiplying scale factor for dependent variable ',
    'depmin': 'F    Minimum value of dependent variable.',
    'depmax': 'F    Maximum value of dependent variable.',
    'depmen': 'F    Mean value of dependent variable.',
    'nzyear': 'N    GMT year corresponding to reference time in file.',
    'nzjday': 'N    GMT julian day.',
    'nzhour': 'N    GMT hour.',
    'nzmin': 'N    GMT minute.',
    'nzsec': 'N    GMT second.',
    'nzmsec': 'N    GMT millisecond.',
    'iztype': '''I    Reference time equivalence:

                  * IUNKN (5): Unknown
                  * IB (9): Begin time
                  * IDAY (10): Midnight of reference GMT day
                  * IO (11): Event origin time
                  * IA (12): First arrival time
                  * ITn (13-22): User defined time pick n, n=0,9''',
    'o': 'F    Event origin time (seconds relative to reference time.)',
    'ko': 'K    Event origin time identification.',
    'a': 'F    First arrival time (seconds relative to reference time.)',
    'ka': 'K    First arrival time identification.',
    'f': 'F    Fini/end of event time (seconds relative to reference time.)',
    'kf': 'F    Fini or end of event time identification.',
    't0': 'F    User defined time (seconds picks or markers relative to '
          'reference time).',
    't1': 'F    User defined time (seconds picks or markers relative to '
          'reference time).',
    't2': 'F    User defined time (seconds picks or markers relative to '
          'reference time).',
    't3': 'F    User defined time (seconds picks or markers relative to '
          'reference time).',
    't4': 'F    User defined time (seconds picks or markers relative to '
          'reference time).',
    't5': 'F    User defined time (seconds picks or markers relative to '
          'reference time).',
    't6': 'F    User defined time (seconds picks or markers relative to '
          'reference time).',
    't7': 'F    User defined time (seconds picks or markers relative to '
          'reference time).',
    't8': 'F    User defined time (seconds picks or markers relative to '
          'reference time).',
    't9': 'F    User defined time (seconds picks or markers relative to '
          'reference time).',
    'kt0': 'F    User defined time pick identification.',
    'kt1': 'F    User defined time pick identification.',
    'kt2': 'F    User defined time pick identification.',
    'kt3': 'F    User defined time pick identification.',
    'kt4': 'F    User defined time pick identification.',
    'kt5': 'F    User defined time pick identification.',
    'kt6': 'F    User defined time pick identification.',
    'kt7': 'F    User defined time pick identification.',
    'kt8': 'F    User defined time pick identification.',
    'kt9': 'F    User defined time pick identification.',
    'kinst': 'K    Generic name of recording instrument',
    'iinst': 'I    Type of recording instrument.',
    'knetwk': 'K    Name of seismic network.',
    'kstnm': 'K    Station name.',
    'istreg': 'I    Station geographic region.',
    'stla': 'F    Station latitude (degrees, north positive)',
    'stlo': 'F    Station longitude (degrees, east positive).',
    'stel': 'F    Station elevation (meters).',
    'stdp': 'F    Station depth below surface (meters).',
    'cmpaz': 'F    Component azimuth (degrees, clockwise from north).',
    'cmpinc': 'F    Component incident angle (degrees, from vertical).',
    'kcmpnm': 'K    Component name.',
    'lpspol': 'L    TRUE if station components have a positive polarity '
              '(left-hand rule).',
    'kevnm': 'K    Event name.',
    'ievreg': 'I    Event geographic region.',
    'evla': 'F    Event latitude (degrees north positive).',
    'evlo': 'F    Event longitude (degrees east positive).',
    'evel': 'F    Event elevation (meters).',
    'evdp': 'F    Event depth below surface (meters).',
    'mag': 'F    Event magnitude.',
    'imagtyp': '''I    Magnitude type:

                  * IMB (52): Bodywave Magnitude
                  * IMS (53): Surfacewave Magnitude
                  * IML (54): Local Magnitude
                  * IMW (55): Moment Magnitude
                  * IMD (56): Duration Magnitude
                  * IMX (57): User Defined Magnitude''',
    'imagsrc': '''I    Source of magnitude information:

                  * INEIC (National Earthquake Information Center)
                  * IPDE (Preliminary Determination of Epicenter)
                  * IISC (International Seismological Centre)
                  * IREB (Reviewed Event Bulletin)
                  * IUSGS (US Geological Survey)
                  * IBRK (UC Berkeley)
                  * ICALTECH (California Institute of Technology)
                  * ILLNL (Lawrence Livermore National Laboratory)
                  * IEVLOC (Event Location (computer program) )
                  * IJSOP (Joint Seismic Observation Program)
                  * IUSER (The individual using SAC2000)
                  * IUNKNOWN (unknown)''',
    'ievtyp': '''I    Type of event:

                  * IUNKN (Unknown)
                  * INUCL (Nuclear event)
                  * IPREN (Nuclear pre-shot event)
                  * IPOSTN (Nuclear post-shot event)
                  * IQUAKE (Earthquake)
                  * IPREQ (Foreshock)
                  * IPOSTQ (Aftershock)
                  * ICHEM (Chemical explosion)
                  * IQB (Quarry or mine blast confirmed by quarry)
                  * IQB1 (Quarry/mine blast with designed shot
                    info-ripple fired)
                  * IQB2 (Quarry/mine blast with observed shot
                    info-ripple fired)
                  * IQMT (Quarry/mining-induced events:
                    tremors and rockbursts)
                  * IEQ (Earthquake)
                  * IEQ1 (Earthquakes in a swarm or aftershock
                    sequence)
                  * IEQ2 (Felt earthquake)
                  * IME (Marine explosion)
                  * IEX (Other explosion)
                  * INU (Nuclear explosion)
                  * INC (Nuclear cavity collapse)
                  * IO_ (Other source of known origin)
                  * IR (Regional event of unknown origin)
                  * IT (Teleseismic event of unknown origin)
                  * IU (Undetermined or conflicting information)
                  * IOTHER (Other)''',
    'nevid': 'N    Event ID (CSS 3.0)',
    'norid': 'N    Origin ID (CSS 3.0)',
    'nwfid': 'N    Waveform ID (CSS 3.0)',
    'khole': 'k    Hole identification if nuclear event.',
    'dist': 'F    Station to event distance (km).',
    'az': 'F    Event to station azimuth (degrees).',
    'baz': 'F    Station to event azimuth (degrees).',
    'gcarc': 'F    Station to event great circle arc length (degrees).',
    'lcalda': 'L    TRUE if DIST AZ BAZ and GCARC are to be calculated '
              'from st event coordinates.',
    'iqual': '''N    Quality of data, as integers. Enum values listed:

                  * IGOOD (45) (Good data)
                  * IGLCH (46) (Glitches)
                  * IDROP (47) (Dropouts)
                  * ILOWSN (48) (Low signal to noise ratio)
                  * IOTHER (44) (Other)''',
    'isynth': '''I    Synthetic data flag:

                  * IRLDTA (Real data)
                  * ????? (Flags for various synthetic seismogram codes)''',
    'user0': 'F    User defined variable storage area 0.',
    'user1': 'F    User defined variable storage area 1.',
    'user2': 'F    User defined variable storage area 2.',
    'user3': 'F    User defined variable storage area 3.',
    'user4': 'F    User defined variable storage area 4.',
    'user5': 'F    User defined variable storage area 5.',
    'user6': 'F    User defined variable storage area 6.',
    'user7': 'F    User defined variable storage area 7.',
    'user8': 'F    User defined variable storage area 8.',
    'user9': 'F    User defined variable storage area 9.',
    'kuser0': 'K    User defined variable storage area 0.',
    'kuser1': 'K    User defined variable storage area 1.',
    'kuser2': 'K    User defined variable storage area 2.',
    'lovrok': 'L    TRUE if it is okay to overwrite this file on disk.'}


# because readable
_hdritems = sorted(DOC.items())
_hdrfmt = "{:10.10s} = {}"
HDR_DESC = '\n'.join([_hdrfmt.format(_hdr, _doc) for _hdr, _doc in _hdritems])

HEADER_DOCSTRING = """
============ ==== ===============================================
Field Name   Type Description
============ ==== ===============================================
""" + HDR_DESC + \
    "\n============ ==== ==============================================="

# Module documentation string
__doc__ = MODULE_DOCSTRING + HEADER_DOCSTRING

# ------------ NULL VALUES ----------------------------------------------------
FNULL = -12345.0
INULL = -12345
SNULL = '-12345  '

# ------------ HEADER NAMES, TYPES, ARRAY POSITIONS ---------------------------
# these are useful b/c they can be used forwards or backwards, like:
# FLOADHDRS.index('az') is 40, and FLOATHDRS[40] is 'az'.
FLOATHDRS = ('delta', 'depmin', 'depmax', 'scale', 'odelta', 'b', 'e', 'o',
             'a', 'internal0', 't0', 't1', 't2', 't3', 't4', 't5', 't6', 't7',
             't8', 't9', 'f', 'resp0', 'resp1', 'resp2', 'resp3', 'resp4',
             'resp5', 'resp6', 'resp7', 'resp8', 'resp9', 'stla', 'stlo',
             'stel', 'stdp', 'evla', 'evlo', 'evel', 'evdp', 'mag', 'user0',
             'user1', 'user2', 'user3', 'user4', 'user5', 'user6', 'user7',
             'user8', 'user9', 'dist', 'az', 'baz', 'gcarc', 'internal1',
             'internal2', 'depmen', 'cmpaz', 'cmpinc', 'xminimum', 'xmaximum',
             'yminimum', 'ymaximum', 'unused6', 'unused7', 'unused8',
             'unused9', 'unused10', 'unused11', 'unused12')

INTHDRS = ('nzyear', 'nzjday', 'nzhour', 'nzmin', 'nzsec', 'nzmsec', 'nvhdr',
           'norid', 'nevid', 'npts', 'internal3', 'nwfid', 'nxsize', 'nysize',
           'unused13', 'iftype', 'idep', 'iztype', 'unused14', 'iinst',
           'istreg', 'ievreg', 'ievtyp', 'iqual', 'isynth', 'imagtyp',
           'imagsrc', 'unused15', 'unused16', 'unused17', 'unused18',
           'unused19', 'unused20', 'unused21', 'unused22', 'leven', 'lpspol',
           'lovrok', 'lcalda', 'unused23')

STRHDRS = ('kstnm', 'kevnm', 'kevnm2', 'khole', 'ko', 'ka', 'kt0', 'kt1',
           'kt2', 'kt3', 'kt4', 'kt5', 'kt6', 'kt7', 'kt8', 'kt9', 'kf',
           'kuser0', 'kuser1', 'kuser2', 'kcmpnm', 'knetwk', 'kdatrd', 'kinst')

# Headers that are in seconds relative to the reference "nz" times
RELHDRS = ('b', 'e', 'a', 'o', 'f', 't0', 't1', 't2', 't3', 't4', 't5', 't6',
           't7', 't8', 't9')

"""
NOTE:

kevnm also has a kevnm2 b/c it takes two array spaces.
'kevnm' lookups must be caught and handled differently.  This happens in the
SACTrace string property getters/setters, .arrayio.dict_to_header_arrays
and .arrayio.header_arrays_to_dict.
"""
# NOTE: using namedtuples for header arrays sounds great, but they're immutable

# TODO: make a dict converter between {'<', '>', '='} and {'little', 'big'}

# ------------ ENUMERATED VALUES ----------------------------------------------
# These are stored in the header as integers.
# Their names and values are given in the mapping below.
# Some (many) are not used.
# TODO: this is ugly; rename things a bit
ENUM_VALS = {'itime': 1, 'irlim': 2, 'iamph': 3, 'ixy': 4, 'iunkn': 5,
             'idisp': 6, 'ivel': 7, 'iacc': 8, 'ib': 9, 'iday': 10, 'io': 11,
             'ia': 12, 'it0': 13, 'it1': 14, 'it2': 15, 'it3': 16, 'it4': 17,
             'it5': 18, 'it6': 19, 'it7': 20, 'it8': 21, 'it9': 22,
             'iradnv': 23, 'itannv': 24, 'iradev': 25, 'itanev': 26,
             'inorth': 27, 'ieast': 28, 'ihorza': 29, 'idown': 30, 'iup': 31,
             'illlbb': 32, 'iwwsn1': 33, 'iwwsn2': 34, 'ihglp': 35, 'isro': 36,
             'inucl': 37, 'ipren': 38, 'ipostn': 39, 'iquake': 40, 'ipreq': 41,
             'ipostq': 42, 'ichem': 43, 'iother': 44, 'igood': 45, 'iglch': 46,
             'idrop': 47, 'ilowsn': 48, 'irldta': 49, 'ivolts': 50, 'imb': 52,
             'ims': 53, 'iml': 54, 'imw': 55, 'imd': 56, 'imx': 57,
             'ineic': 58, 'ipdeq': 59, 'ipdew': 60, 'ipde': 61, 'iisc': 62,
             'ireb': 63, 'iusgs': 64, 'ibrk': 65, 'icaltech': 66, 'illnl': 67,
             'ievloc': 68, 'ijsop': 69, 'iuser': 70, 'iunknown': 71, 'iqb': 72,
             'iqb1': 73, 'iqb2': 74, 'iqbx': 75, 'iqmt': 76, 'ieq': 77,
             'ieq1': 78, 'ieq2': 79, 'ime': 80, 'iex': 81, 'inu': 82,
             'inc': 83, 'io_': 84, 'il': 85, 'ir': 86, 'it': 87, 'iu': 88,
             'ieq3': 89, 'ieq0': 90, 'iex0': 91, 'iqc': 92, 'iqb0': 93,
             'igey': 94, 'ilit': 95, 'imet': 96, 'iodor': 97, 'ios': 103}

# reverse look-up: you have the number, want the string
ENUM_NAMES = dict((v, k) for k, v in ENUM_VALS.items())


# accepted values, by header
ACCEPTED_VALS = {'iftype': ['itime', 'irlim', 'iamph', 'ixy'],
                 'idep': ['iunkn', 'idisp', 'ivel', 'ivolts', 'iacc'],
                 'iztype': ['iunkn', 'ib', 'iday', 'io', 'ia', 'it0', 'it1',
                            'it2', 'it3', 'it4', 'it5', 'it6', 'it7', 'it8',
                            'it9'],
                 'imagtyp': ['imb', 'ims', 'iml', 'imw', 'imd', 'imx'],
                 'imagsrc': ['ineic', 'ipde', 'iisc', 'ireb', 'iusgs', 'ipdeq',
                             'ibrk', 'icaltech', 'illnl', 'ievloc', 'ijsop',
                             'iuser', 'iunknown'],
                 'ievtyp': ['iunkn', 'inucl', 'ipren', 'ipostn', 'iquake',
                            'ipreq', 'ipostq', 'ichem', 'iqb', 'iqb1', 'iqb2',
                            'iqbx', 'iqmt', 'ieq', 'ieq1', 'ieq2', 'ime',
                            'iex', 'inu', 'inc', 'io_', 'il', 'ir', 'it', 'iu',
                            'iother'],
                 'iqual': ['igood', 'iglch', 'idrop', 'ilowsn', 'iother'],
                 'isynth': ['irldta']}

ACCEPTED_INT = ACCEPTED_VALS.copy()
for _hdr in ACCEPTED_INT:
    ACCEPTED_INT[_hdr] = [ENUM_VALS[_ival] for _ival in ACCEPTED_VALS[_hdr]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
seiscomp.event test suite.

:author:
    EOST (École et Observatoire des Sciences de la Terre)
:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import filecmp
import os
import re

from lxml import etree

from obspy import read_events
from obspy.core.util.base import NamedTemporaryFile
from obspy.io.quakeml.core import _read_quakeml
from obspy.io.quakeml.core import _validate as _validate_quakeml
from obspy.io.seiscomp.core import validate as validate_scml
from obspy.io.seiscomp.event import (
    SCHEMA_VERSION, NEW_SCHEMA_VERSION, _read_scml
)
import pytest


class TestEvent():
    """
    Test suite for obspy.io.seiscomp.event
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, root, datapath):
        # directory where the test files are located
        self.io_directory = root / 'io'
        self.quakeml_path = self.io_directory / 'quakeml' / 'tests' / 'data'
        self.path = datapath
        self.write_xslt_filename = (
            self.io_directory / 'seiscomp' / 'data' /
            'quakeml_1.2__sc3ml_0.10.xsl')
        self.schema_pattern = re.compile(
            r'http://geofon\.gfz(?:-potsdam)?\.de/ns/seiscomp3?-schema/[\d\.]+'
        )
        self.version_pattern = re.compile(r'version="[\d\.]+"')

    def change_reading_version(self, filename, version):
        """
        Change the namespace and version number of a SCML
        file-like object and return an etree document.
        """
        with open(filename, 'r') as f:
            data = f.read()

            # Remove encoding declaration otherwise lxml will not be
            # able to read the file.
            data = data.replace('<?xml version="1.0" encoding="UTF-8"?>\n', '')

            new_ns = "http://geofon.gfz.de/ns/seiscomp-schema/"
            old_ns = "http://geofon.gfz-potsdam.de/ns/seiscomp3-schema/"

            if version in NEW_SCHEMA_VERSION:
                sch_ver = f"{new_ns}{version}"
                data = data.replace(old_ns, new_ns)
            else:
                sch_ver = f"{old_ns}{version}"
                data = data.replace(new_ns, old_ns)

            data = self.schema_pattern.sub(sch_ver, data)
            data = self.version_pattern.sub(
                f'version="{version}"', data)

        return etree.fromstring(data)

    def change_writing_version(self, data, version):
        """
        Change the namespace and version number of a
        string and return the new string.
        """
        # Remove the XML header else the version will be changed
        data = data.replace('<?xml version="1.0" encoding="UTF-8"?>\n', '')

        new_ns = "http://geofon.gfz.de/ns/seiscomp-schema/"
        old_ns = "http://geofon.gfz-potsdam.de/ns/seiscomp3-schema/"

        if version in NEW_SCHEMA_VERSION:
            sch_ver = new_ns + version
        else:
            sch_ver = old_ns + version

        data = self.schema_pattern.sub(sch_ver, data)
        data = self.version_pattern.sub(
            f'version="{version}"', data)

        # Put back the XML header
        return '<?xml version="1.0" encoding="UTF-8"?>\n%s' % data

    def cmp_read_xslt_file(self, scml_file, quakeml_file, validate=True):
        """
        Check if the QuakeML file generated with the XSLT file is the
        same than the one in the data folder. Every available SCML
        versions are tested except those for which the file is not
        valid. Test file namespaces are briefly hacked to match
        variable incoming SCML.
        """
        for version in SCHEMA_VERSION:
            # versions 0.9 and below still have incorrect km - m
            # ellipsoid conversion but are kept for legacy reasons.
            # accordingly they no longer pass _origin tests
            if version in ['0.7', '0.8', '0.9']:
                continue

            if version in NEW_SCHEMA_VERSION:
                xslfile = 'scml_%s__quakeml_1.2.xsl' % version
            else:
                xslfile = 'sc3ml_%s__quakeml_1.2.xsl' % version

            read_xslt_filename = os.path.join(
                self.io_directory, 'seiscomp', 'data',
                xslfile,
            )

            transform = etree.XSLT(etree.parse(read_xslt_filename))
            filename = self.path / scml_file
            scml_doc = self.change_reading_version(filename, version)

            if not validate_scml(scml_doc):
                continue

            quakeml_doc = transform(scml_doc)

            with NamedTemporaryFile() as tf:
                tf.write(quakeml_doc)
                if validate:
                    assert _validate_quakeml(tf.name)

                # Read the expected file and force matching namespace
                filepath_cmp = self.path / quakeml_file
                with open(filepath_cmp, 'r') as f:
                    expected = f.read()

                new_ns = "gfz.de"
                old_ns = "gfz-potsdam.de"

                if version in NEW_SCHEMA_VERSION:
                    expected = expected.replace(old_ns, new_ns)
                else:
                    expected = expected.replace(new_ns, old_ns)

                # Read generated content
                tf.seek(0)
                generated = tf.read().decode('utf-8')
                print("DEBUG:: ", version)

                assert generated == expected

    def cmp_write_xslt_file(self, quakeml_file, scml_file, target_version,
                            validate=True, path=None):
        """
        Check if the SCML file generated with the XSLT file is the
        same than the one in the data folder.

        The `target_version` parameter allows to change the version of
        the generated file to match the version of the reference file.
        """
        if path is None:
            path = self.path

        transform = etree.XSLT(etree.parse(self.write_xslt_filename))
        filename = os.path.join(path, quakeml_file)
        quakeml_doc = etree.parse(filename)
        scml_doc = transform(quakeml_doc)

        scml_doc_string = self.change_writing_version(
            bytes(scml_doc).decode(encoding=quakeml_doc.docinfo.encoding),
            target_version,
        )

        scml_doc_bytes = scml_doc_string.encode(
            encoding=quakeml_doc.docinfo.encoding
        )

        with NamedTemporaryFile() as tf:
            tf.write(scml_doc_bytes)

            if validate:
                assert validate_scml(tf.name, target_version)
            filepath_cmp = self.path / scml_file
            assert filecmp.cmp(filepath_cmp, tf.name)

    def test_scml_versions(self, testdata):
        """
        Test multiple schema versions
        """
        for version in ['0.9', '0.10', '0.11',
                        '0.12', '0.13', '0.14']:
            filename = testdata['version%s' % version]
            read_events(filename)

        filename = testdata['version0.6']
        expected_message = re.escape(
            "Can't read SCML version 0.6, ObsPy can handle "
            "versions [0.7, 0.8, 0.9, 0.10, "
            "0.11, 0.12, 0.13, 0.14].")
        with pytest.raises(ValueError, match=expected_message):
            read_events(filename)

        filename = testdata['version0.15']
        expected_message = re.escape(
            "Can't read SCML version 0.15, ObsPy can handle "
            "versions [0.7, 0.8, 0.9, 0.10, "
            "0.11, 0.12, 0.13, 0.14].")
        with pytest.raises(ValueError, match=expected_message):
            read_events(filename)

    def test_read_xslt_event(self):
        self.cmp_read_xslt_file('quakeml_1.2_event.sc3ml',
                                'quakeml_1.2_event_res.xml')

    def test_read_xslt_origin(self):
        self.cmp_read_xslt_file('quakeml_1.2_origin.sc3ml',
                                'quakeml_1.2_origin_res.xml')

    def test_read_xslt_magnitude(self):
        self.cmp_read_xslt_file('quakeml_1.2_magnitude.sc3ml',
                                'quakeml_1.2_magnitude.xml')

    def test_read_xslt_station_magnitude_contribution(self):
        self.cmp_read_xslt_file(
            'quakeml_1.2_stationmagnitudecontributions.sc3ml',
            'quakeml_1.2_stationmagnitudecontributions.xml')

    def test_read_xslt_station_magnitude(self):
        self.cmp_read_xslt_file('quakeml_1.2_stationmagnitude.sc3ml',
                                'quakeml_1.2_stationmagnitude.xml')

    def test_read_xslt_data_used_in_moment_tensor(self):
        self.cmp_read_xslt_file('quakeml_1.2_data_used.sc3ml',
                                'quakeml_1.2_data_used.xml')

    def test_read_xslt_arrival(self):
        self.cmp_read_xslt_file('quakeml_1.2_arrival.sc3ml',
                                'quakeml_1.2_arrival.xml')

    def test_read_xslt_pick(self):
        self.cmp_read_xslt_file('quakeml_1.2_pick.sc3ml',
                                'quakeml_1.2_pick.xml')

    def test_read_xslt_focalmechanism(self):
        self.cmp_read_xslt_file('quakeml_1.2_focalmechanism.sc3ml',
                                'quakeml_1.2_focalmechanism_res.xml')

    def test_read_xslt_amplitude(self):
        """ See issue #2273 """
        self.cmp_read_xslt_file('quakeml_1.2_amplitude.sc3ml',
                                'quakeml_1.2_amplitude_res.xml')

    def test_read_xslt_iris_events(self):
        # Magnitude lost during conversion
        self.cmp_read_xslt_file('iris_events.sc3ml', 'iris_events_res.xml')

    def test_read_xslt_example(self):
        self.cmp_read_xslt_file('qml-example-1.2-RC3.sc3ml',
                                'qml-example-1.2-RC3.xml')

    def test_read_example(self, testdata):
        filename = testdata['qml-example-1.2-RC3.sc3ml']
        catalog = _read_scml(filename)

        assert len(catalog.events) == 1
        assert len(catalog.events[0].origins) == 1

    def test_read_id_prefix(self, testdata):
        filename = \
            testdata['qml-example-1.2-RC3_wrong_id.sc3ml']
        catalog = _read_scml(filename, id_prefix='quakeml:obspy.org/')

        assert len(catalog.events) == 1
        event = catalog.events[0]
        assert event.resource_id == 'quakeml:obspy.org/test_event_id'

        assert len(event.origins) == 1
        origin = event.origins[0]
        assert origin.resource_id == 'quakeml:obspy.org/test_origin_id'
        assert origin.reference_system_id == \
            'quakeml:obspy.org/test_reference_system_id'

        assert len(event.amplitudes) == 1
        amplitude = event.amplitudes[0]
        assert amplitude.resource_id == 'quakeml:obspy.org/test_amplitude_id'

        assert len(event.magnitudes) == 1
        magnitude = event.magnitudes[0]
        assert magnitude.resource_id == 'quakeml:obspy.org/test_magnitude_id'

        assert len(event.station_magnitudes) == 1
        station_magnitude = event.station_magnitudes[0]
        assert station_magnitude.resource_id == \
            'quakeml:obspy.org/test_station_magnitude_id'

    def test_read_string(self, testdata):
        """
        Test reading a SCML string/unicode object via read_events.
        """
        filename = testdata['qml-example-1.2-RC3.sc3ml']
        with open(filename, 'rb') as fp:
            data = fp.read()

            catalog = read_events(data)
            assert len(catalog) == 1

    def test_read_quakeml(self, testdata):
        """
        Test reading a QuakeML file via read_events.
        """
        filename = testdata['qml-example-1.2-RC3.xml']
        expected_message = "Not a SCML compatible file or string."
        with pytest.raises(ValueError, match=expected_message):
            read_events(filename, format='SCML')

    def test_read_scml_fields(self, testdata):
        """
        Test SCML fields which are not in the QuakeML 1.2.
        """
        filename = testdata['sc3ml_fields.sc3ml']
        catalog = read_events(filename, format='SCML')

        # No longer present
        # assert 'reading' in catalog.extra

        origin = catalog[0].origins[0]
        assert 'pdf' in origin.time_errors.extra
        assert 'modificationTime' in origin.creation_info.extra

        comment = origin.comments[0]
        assert 'start' in comment.extra
        assert 'end' in comment.extra

        station_magnitude = catalog[0].station_magnitudes[0]
        assert 'passedQC' in station_magnitude.extra

        tensor = catalog[0].focal_mechanisms[0].moment_tensor
        assert 'status' in tensor.extra
        assert 'cmtName' in tensor.extra
        assert 'cmtVersion' in tensor.extra
        assert 'phaseSetting' in tensor.extra
        assert 'stationMomentTensorContribution' in tensor.extra

    def test_write_xslt_event(self):
        self.cmp_write_xslt_file('quakeml_1.2_event.xml',
                                 'quakeml_1.2_event.sc3ml',
                                 '0.10',
                                 path=self.quakeml_path)

    def test_write_xslt_origin(self):
        self.cmp_write_xslt_file('quakeml_1.2_origin.xml',
                                 'quakeml_1.2_origin.sc3ml',
                                 '0.12',
                                 path=self.quakeml_path)

    def test_write_xslt_quarryblast(self):
        self.cmp_write_xslt_file('quakeml_1.2_quarryblast.xml',
                                 'quakeml_1.2_quarryblast.scml',
                                 '0.14')

    def test_write_xslt_magnitude(self):
        # Missing origin in original QuakeML test case.
        self.cmp_write_xslt_file('quakeml_1.2_magnitude.xml',
                                 'quakeml_1.2_magnitude.sc3ml',
                                 '0.12')

    def test_write_xslt_station_magnitude_contribution(self):
        # Missing origin in original QuakeML test case.
        self.cmp_write_xslt_file(
            'quakeml_1.2_stationmagnitudecontributions.xml',
            'quakeml_1.2_stationmagnitudecontributions.sc3ml',
            '0.12',
        )

    def test_write_xslt_station_magnitude(self):
        # Missing origin in original QuakeML test case.
        self.cmp_write_xslt_file('quakeml_1.2_stationmagnitude.xml',
                                 'quakeml_1.2_stationmagnitude.sc3ml',
                                 '0.10')

    def test_write_xslt_data_used_in_moment_tensor(self):
        self.cmp_write_xslt_file('quakeml_1.2_data_used.xml',
                                 'quakeml_1.2_data_used.scml',
                                 '0.14')

    def test_write_xslt_arrival(self):
        self.cmp_write_xslt_file('quakeml_1.2_arrival.xml',
                                 'quakeml_1.2_arrival_res.sc3ml',
                                 '0.13')

    def test_write_xslt_pick(self):
        self.cmp_write_xslt_file('quakeml_1.2_pick.xml',
                                 'quakeml_1.2_pick.sc3ml',
                                 '0.10')

    def test_write_xslt_012_events(self):
        self.cmp_write_xslt_file('westaus_events.xml',
                                 'westaus_events_0.12',
                                 '0.12')

    def test_write_xslt_014_events(self):
        self.cmp_write_xslt_file('westaus_events.xml',
                                 'westaus_events_0.14',
                                 '0.14')

    def test_write_xslt_focalmechanism(self):
        self.cmp_write_xslt_file('quakeml_1.2_focalmechanism.xml',
                                 'quakeml_1.2_focalmechanism.sc3ml',
                                 '0.11',
                                 path=self.quakeml_path)

    def test_write_xslt_iris_events(self):
        self.cmp_write_xslt_file('iris_events.xml',
                                 'iris_events.sc3ml',
                                 '0.10',
                                 path=self.quakeml_path)

    def test_write_xslt_neries_events(self):
        # Some ID are generated automatically. File comparison can't be done.
        filename = os.path.join(self.quakeml_path, 'neries_events.xml')
        catalog = _read_quakeml(filename)

        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            try:
                catalog.write(tmpfile, format='SCML', validate=True,
                              verbose=True)
            except AssertionError as e:
                self.fail(e)

    def test_write_xslt_usgs_events(self):
        # Can't validate due to wrong event types (QuakeML was already
        # unvalid)
        self.cmp_write_xslt_file('usgs_event.xml',
                                 'usgs_event.sc3ml',
                                 '0.11',
                                 path=self.quakeml_path,
                                 validate=False)

    def test_write_xslt_example(self):
        self.cmp_write_xslt_file('qml-example-1.2-RC3.xml',
                                 'qml-example-1.2-RC3.sc3ml',
                                 '0.11')

    def test_write_example(self, testdata):
        filename = testdata['qml-example-1.2-RC3.xml']
        catalog = _read_quakeml(filename)

        with NamedTemporaryFile() as tf:
            catalog.write(tf, format='SCML', validate=True,
                          version='0.13')
            filepath_cmp = \
                testdata['qml-example-1.2-RC3_write.sc3ml']
            assert filecmp.cmp(filepath_cmp, tf.name)

    def test_write_remove_events(self, testdata):
        filename = testdata['qml-example-1.2-RC3.xml']
        catalog = _read_quakeml(filename)

        with NamedTemporaryFile() as tf:
            catalog.write(tf, format='SCML', validate=True,
                          event_removal=True, version='0.13')
            filepath_cmp = \
                testdata['qml-example-1.2-RC3_no_events.sc3ml']
            assert filecmp.cmp(filepath_cmp, tf.name)

    def test_read_and_write(self, testdata):
        filename = testdata['qml-example-1.2-RC3_write.sc3ml']
        catalog = read_events(filename)

        with NamedTemporaryFile() as tf:
            catalog.write(tf, format='SCML', validate=True,
                          version='0.13')
            assert filecmp.cmp(filename, tf.name)

    def test_read_and_write_scml_fields(self, testdata):
        """
        Test that the non-QuakeML nodes are correctly rewritten.
        """
        filename = testdata['scml_fields.scml']
        catalog = read_events(filename)

        with NamedTemporaryFile() as tf:
            catalog.write(tf, format='SCML', validate=True,
                          version='0.14')
            assert filecmp.cmp(filename, tf.name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import warnings

from obspy import UTCDateTime, read_events
from obspy.io.ndk.core import (ObsPyNDKException, _parse_date_time, _is_ndk,
                               _read_ndk)
import pytest


class TestNDK():
    """
    Test suite for obspy.io.ndk
    """
    def test_read_single_ndk(self, testdata):
        """
        Test reading a single event from an NDK file and comparing it to a
        QuakeML file that has been manually checked to contain all the
        information in the NDK file.
        """
        filename = testdata['C200604092050A.ndk']
        cat = _read_ndk(filename)

        reference = testdata['C200604092050A.xml']
        ref_cat = read_events(reference)

        assert cat == ref_cat

    def test_read_multiple_events(self, testdata):
        """
        Tests the reading of multiple events in one file. The file has been
        edited to test a variety of settings.
        """
        filename = testdata['multiple_events.ndk']
        cat = _read_ndk(filename)

        assert len(cat) == 6

        # Test the type of moment tensor inverted for.
        assert [
            i.focal_mechanisms[0].moment_tensor.inversion_type
            for i in cat] == ["general", "zero trace", "double couple"] * 2

        # Test the type and duration of the moment rate function.
        assert [
            i.focal_mechanisms[0].moment_tensor.source_time_function.type
            for i in cat] == ["triangle", "box car"] * 3
        assert [
            i.focal_mechanisms[0].moment_tensor.source_time_function.duration
            for i in cat] == [2.6, 7.4, 9.0, 1.8, 2.0, 1.6]

        # Test the type of depth setting.
        assert [i.preferred_origin().depth_type for i in cat] == \
            ["from moment tensor inversion", "from location",
             "from modeling of broad-band P waveforms"] * 2

        # Check the solution type.
        for event in cat[:3]:
            assert "Standard" in \
                          event.focal_mechanisms[0].comments[0].text
        for event in cat[3:]:
            assert "Quick" in \
                          event.focal_mechanisms[0].comments[0].text

    def test_is_ndk(self, testdata, datapath):
        """
        Test for the the _is_ndk() function.
        """
        valid_files = [testdata['C200604092050A.ndk'],
                       testdata['multiple_events.ndk']]
        invalid_files = list(datapath.parent.glob('*.py'))
        assert len(invalid_files) > 0

        for filename in valid_files:
            assert _is_ndk(filename)
        for filename in invalid_files:
            assert not _is_ndk(filename)

    def test_reading_using_obspy_plugin(self, testdata):
        """
        Checks that reading with the read_events() function works correctly.
        """
        filename = testdata['C200604092050A.ndk']
        cat = read_events(filename)

        reference = testdata['C200604092050A.xml']
        ref_cat = read_events(reference)

        assert cat == ref_cat

    def test_reading_from_string_io(self, testdata):
        """
        Tests reading from StringIO.
        """
        filename = testdata['C200604092050A.ndk']
        with open(filename, "rt") as fh:
            file_object = io.StringIO(fh.read())

        cat = read_events(file_object)
        file_object.close()

        reference = testdata['C200604092050A.xml']
        ref_cat = read_events(reference)

        assert cat == ref_cat

    def test_reading_from_bytes_io(self, testdata):
        """
        Tests reading from BytesIO.
        """
        filename = testdata['C200604092050A.ndk']
        with open(filename, "rb") as fh:
            file_object = io.BytesIO(fh.read())

        cat = read_events(file_object)
        file_object.close()

        reference = testdata['C200604092050A.xml']
        ref_cat = read_events(reference)

        assert cat == ref_cat

    def test_reading_from_open_file_in_text_mode(self, testdata):
        """
        Tests reading from an open file in text mode.
        """
        filename = testdata['C200604092050A.ndk']
        with open(filename, "rt") as fh:
            cat = read_events(fh)

        reference = testdata['C200604092050A.xml']
        ref_cat = read_events(reference)

        assert cat == ref_cat

    def test_reading_from_open_file_in_binary_mode(self, testdata):
        """
        Tests reading from an open file in binary mode.
        """
        filename = testdata['C200604092050A.ndk']
        with open(filename, "rb") as fh:
            cat = read_events(fh)

        reference = testdata['C200604092050A.xml']
        ref_cat = read_events(reference)

        assert cat == ref_cat

    def test_reading_the_same_file_twice_does_not_raise_a_warnings(
            self, testdata):
        """
        Asserts that reading the same file twice does not raise a warning
        due to resource identifier already in use.
        """
        filename = testdata['C200604092050A.ndk']
        cat_1 = read_events(filename)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cat_2 = read_events(filename)

        assert len(w) == 0
        assert cat_1 == cat_2

        filename = testdata['multiple_events.ndk']
        cat_1 = read_events(filename)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cat_2 = read_events(filename)

        assert len(w) == 0
        assert cat_1 == cat_2

    def test_is_ndk_for_file_with_invalid_date(self, testdata):
        """
        Tests the _is_ndk function for a file with invalid date.
        """
        assert not _is_ndk(testdata["faulty_invalid_date.ndk"])

    def test_is_ndk_for_file_with_invalid_latitude(self, testdata):
        """
        Tests the _is_ndk function a file with an invalid latitude.
        """
        assert not _is_ndk(testdata["faulty_invalid_latitude.ndk"])

    def test_is_ndk_for_file_with_infeasible_latitude(self, testdata):
        """
        Tests the _is_ndk function a file with an unfeasible latitude.
        """
        assert not _is_ndk(testdata["faulty_infeasible_latitude.ndk"])

    def test_reading_file_with_multiple_errors(self, testdata):
        """
        Tests reading a file with multiple errors.
        """
        filename = testdata['faulty_multiple_events.ndk']

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cat = read_events(filename)

        assert len(w) == 6
        assert "Invalid time in event 2" in str(w[0])
        assert "Unknown data type" in str(w[1])
        assert "Moment rate function" in str(w[2])
        assert "Unknown source type" in str(w[3])
        assert "Unknown type of depth" in str(w[4])
        assert "Invalid CMT timestamp" in str(w[5])

        # One event should still be available.
        assert len(cat) == 1

    def test_reading_from_string(self, testdata):
        """
        Tests reading from a string.
        """
        filename = testdata['C200604092050A.ndk']

        reference = testdata['C200604092050A.xml']
        ref_cat = read_events(reference)

        with io.open(filename, "rt") as fh:
            data = fh.read()

        assert _is_ndk(data)
        cat = _read_ndk(data)

        assert cat == ref_cat

    def test_reading_from_bytestring(self, testdata):
        """
        Tests reading from a byte string.
        """
        filename = testdata['C200604092050A.ndk']

        reference = testdata['C200604092050A.xml']
        ref_cat = read_events(reference)

        with io.open(filename, "rb") as fh:
            data = fh.read()

        assert _is_ndk(data)
        cat = _read_ndk(data)

        assert cat == ref_cat

    def test_missing_lines(self, testdata):
        """
        Tests the raised warning if an event has less then 5 lines.
        """
        with open(testdata['multiple_events.ndk'], "rt") \
                as fh:
            lines = [_i.rstrip() for _i in fh.readlines()]

        # Assemble anew and skip last line.
        data = io.StringIO("\n".join(lines[:-1]))

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cat = read_events(data)

        data.close()

        assert len(w) == 1
        assert "Not a multiple of 5 lines" in str(w[0])
        # Only five events will have been read.
        assert len(cat) == 5

    def test_reading_event_with_faulty_but_often_occurring_timestamp(
            self, testdata):
        """
        The timestamp "O-00000000000000" is not valid according to the NDK
        definition but occurs a lot in the GCMT catalog thus we include it
        here.
        """
        filename = testdata['faulty_cmt_timestamp.ndk']

        cat = read_events(filename)

        assert len(cat) == 1
        comments = cat[0].focal_mechanisms[0].comments
        assert "CMT Analysis Type: Unknown" in comments[0].text
        assert "CMT Timestamp: O-000000000" in comments[1].text

    def test_raise_exception_if_no_events_in_file(self, testdata):
        """
        The parser is fairly relaxed and will skip invalid files. This test
        assures that an exception is raised if every event has been skipped.
        """
        with open(testdata['C200604092050A.ndk'], "rt") \
                as fh:
            lines = [_i.rstrip() for _i in fh.readlines()]

        # Assemble anew and skip last line.
        data = io.StringIO("\n".join(lines[:-1]))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with pytest.raises(ObsPyNDKException):
                read_events(data)

    def test_parse_date_time_function(self):
        """
        Tests the _parse_date_time() function.
        """
        # Simple tests for some valid times.
        date, time = "1997/11/03", "19:17:33.8"
        assert _parse_date_time(date, time) == \
            UTCDateTime(1997, 11, 3, 19, 17, 33, int(8E5))
        date, time = "1996/11/20", "19:42:56.1"
        assert _parse_date_time(date, time) == \
            UTCDateTime(1996, 11, 20, 19, 42, 56, int(1E5))
        date, time = "2005/01/01", "01:20:05.4"
        assert _parse_date_time(date, time) == \
            UTCDateTime(2005, 1, 1, 1, 20, 5, int(4E5))
        date, time = "2013/03/01", "03:29:46.8"
        assert _parse_date_time(date, time) == \
            UTCDateTime(2013, 3, 1, 3, 29, 46, int(8E5))
        date, time = "2013/03/02", "07:53:43.8"
        assert _parse_date_time(date, time) == \
            UTCDateTime(2013, 3, 2, 7, 53, 43, int(8E5))

        # Some more tests for 60s. The tested values are all values occurring
        # in a big NDK test file.
        date, time = "1998/09/27", "00:57:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(1998, 9, 27, 0, 58)
        date, time = "2000/12/22", "16:29:60.0"
        assert _parse_date_time(date, time) == \
            UTCDateTime(2000, 12, 22, 16, 30)
        date, time = "2003/06/19", "23:04:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2003, 6, 19, 23, 5)
        date, time = "2005/06/20", "02:32:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2005, 6, 20, 2, 33)
        date, time = "2006/03/02", "17:16:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2006, 3, 2, 17, 17)
        date, time = "2006/05/26", "10:25:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2006, 5, 26, 10, 26)
        date, time = "2006/08/20", "13:34:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2006, 8, 20, 13, 35)
        date, time = "2007/04/20", "00:30:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2007, 4, 20, 0, 31)
        date, time = "2007/07/02", "00:54:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2007, 7, 2, 0, 55)
        date, time = "2007/08/27", "17:11:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2007, 8, 27, 17, 12)
        date, time = "2008/09/24", "01:36:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2008, 9, 24, 1, 37)
        date, time = "2008/10/05", "10:44:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2008, 10, 5, 10, 45)
        date, time = "2009/04/17", "04:09:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2009, 4, 17, 4, 10)
        date, time = "2009/06/03", "14:30:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2009, 6, 3, 14, 31)
        date, time = "2009/07/20", "10:44:60.0"
        assert _parse_date_time(date, time) == UTCDateTime(2009, 7, 20, 10, 45)



"""
.. note::
    This module can be easily extended to write and read more event
    information.
    If you are interested, please send a PR to the github repository.
      1. Add 'extended' or similar key to FIELDS dict, e.g. as a start use
          'extended': (
              '{time!s:.25} {lat:.6f} {lat_err:.6f} {lon:.6f} {lon_err:.6f} '
              '{dep:.3f} {dep_err:.3f} {mag:.2f} {mag_err:.2f} {magtype} {id}'
              )
         You can also test by just passing the string to `fields` option.
      2. Implement writing functionality in write_csv by adding the new
         properties to the dict d. Missing values have to be handled.
      3. Implement reading functionality for the new properties in read_csv.
      4. Write some tests testing all new properties. Check that an event with
         all new properties defined or missing can be written and read again.
    Similar can be done for the picks using PFIELDS, _write_picks, _read_picks
"""

import csv
from contextlib import contextmanager
import io
import math
import pathlib
from string import Formatter
from warnings import warn
import zipfile


import numpy as np
from obspy import UTCDateTime as UTC, __version__
from obspy.core import event as evmod
from obspy.core.util.decorator import map_example_filename


# for reading
DEFAULT = {'magtype': ''}
# for writing
FIELDS = {
    'basic': (
        '{id} {time!s:.25} {lat:.6f} {lon:.6f} {dep:.3f} '
        '{magtype} {mag:.2f}'),
    'eventtxt': (
        '{id},{time!s:.25},{lat:.6f},{lon:.6f},{dep:.3f},{author},,{contrib},,'
        '{magtype},{mag:.2f},{magauthor},{region}').split(',')
}
PFIELDS = {
    'basic': '{seedid} {phase} {time:.5f} {weight:.3f}'
}
CSZ_COMMENT = f'CSZ format ObsPy v{__version__} obspy_no_uncompress'.encode(
    'utf-8')
# for load_csv
DTYPE = {
    'time': 'datetime64[ms]',
    'lat': float,
    'lon': float,
    'dep': float,
    'mag': float,
    'magtype': 'U10',
    'id': 'U50'
}

# catalog and contribid are not used
EVENTTXT_NAMES = (  # for reading
    'id time lat lon dep author catalog contrib contribid '
    'magtype mag magauthor region')
EVENTTXT_HEADER = (  # for writing
    '#EventID | Time | Latitude | Longitude | Depth/km | Author | Catalog | '
    'Contributor | ContributorID | MagType | Magnitude | MagAuthor | '
    'EventLocationName')


def _is_csv(fname, **kwargs):
    try:
        return _read_csv(fname, format_check=True)
    except Exception:
        return False


def _is_eventtxt(fname, **kwargs):
    try:
        return _read_eventtxt(fname, format_check=True)
    except Exception:
        return False


def _is_csz(fname, **kwargs):
    try:
        if not zipfile.is_zipfile(fname):
            return False
        with zipfile.ZipFile(fname) as zipf:
            if not (zipf.comment.startswith(b'CSZ') and
                    zipf.comment.endswith(b'obspy_no_uncompress')):
                return False
        return True
    except Exception:
        return False


def _evid(event):
    return str(event.resource_id).split('/')[-1]


def _origin(event):
    return event.preferred_origin() or event.origins[0]


@contextmanager
def _open(filein, *args, **kwargs):
    """Accept files or file names or pathlib objects"""
    if isinstance(filein, (str, pathlib.PurePath)):  # filename
        with open(filein, *args, **kwargs) as f:
            yield f
    else:  # file-like object
        yield filein


def _names_sequence(names):
    if isinstance(names, dict):
        names = [names.get(i, '_') for i in range(max(names.keys())+1)]
    elif ' ' in names:
        names = names.split()
    return names


def _string(row, key):
    if key in row and row[key].strip() != '':
        return row[key].strip()


def _read_csv(fname, skipheader=0, default=None, names=None,
              format_check=False, **kwargs):
    """
    Read a CSV file and return ObsPy catalog

    :param fname: filename or file-like object to read from
    :param skipheader: skip first rows of file
    :param default: dictionary with default values, at the moment only
         magtype is supported,
         i.e. to set magtypes use ``default={'magtype': 'Ml'}``
    :param names: determined automatically from header line of file,
        otherwise can be specified as string, sequence or dict
    :param format_check: Check if the first event can be read
    :param **kargs: all other kwargs are passed to csv.DictReader,
        important additional arguments are fieldnames, dialect, delimiter, etc

    Example
    -------

    >>> from obspy import read_events
    >>> events = read_events('/path/to/catalog.csv')
    >>> print(events)  # doctest: +NORMALIZE_WHITESPACE
    3 Event(s) in Catalog:
    2012-04-04T14:21:42.300000Z | +41.818,  +79.689 | 4.4  mb
    2012-04-04T14:18:37.000000Z | +39.342,  +41.044 | 4.3  ML
    2012-04-04T14:08:46.000000Z | +38.017,  +37.736 | 3.0  ML

    You can read CSV files created by this module or external CSV files.
    Example reading an external CSV file:

    >>> from obspy import read_events
    >>> names = 'year mon day hour minu sec _ lat lon dep mag id'
    >>> events = read_events(
    ...     '/path/to/external.csv', 'CSV', skipheader=1, names=names)
    >>> print(events)  # doctest: +NORMALIZE_WHITESPACE
    1 Event(s) in Catalog:
    2023-05-06T19:55:01.300000Z | +10.194, +124.830 | 0.2
    """
    if default is None:
        default = DEFAULT
    events = []
    with _open(fname) as f:
        for _ in range(skipheader):
            f.readline()
        if names is not None:
            kwargs.setdefault('fieldnames', _names_sequence(names))
        reader = csv.DictReader(f, **kwargs)
        for row in reader:
            if 'time' in row:
                time = UTC(row['time'])
            else:
                time = UTC(
                    '{year}-{mon}-{day} {hour}:{minu}:{sec}'.format(**row))
            try:
                if 'depm' in row:
                    dep = float(row['depm'])
                else:
                    dep = float(row['dep']) * 1000
                if math.isnan(dep):
                    raise
            except Exception:
                dep = None
            author = _string(row, 'author')
            contrib = _string(row, 'contrib')
            if author is not None or contrib is not None:
                info = evmod.CreationInfo(author=author, agency_id=contrib)
            else:
                info = None
            origin = evmod.Origin(
                time=time,
                latitude=row['lat'],
                longitude=row['lon'],
                depth=dep,
                creation_info=info
            )
            try:
                # add zero to eliminate negative zeros in magnitudes
                mag = float(row['mag']) + 0
                if math.isnan(mag):
                    raise
            except Exception:
                magnitudes = []
            else:
                try:
                    magtype = row['magtype']
                    if magtype.lower() in ('', 'none', 'null', 'nan'):
                        raise
                except Exception:
                    magtype = default.get('magtype')
                magauthor = _string(row, 'magauthor')
                info = (evmod.CreationInfo(author=magauthor) if magauthor
                        else None)
                magnitudes = [evmod.Magnitude(
                    mag=mag, magnitude_type=magtype, creation_info=info)]
            try:
                id_ = evmod.ResourceIdentifier(row['id'].strip())
            except Exception:
                id_ = None
            region = _string(row, 'region')
            descs = ([evmod.EventDescription(region, 'region name')]
                     if region else [])
            event = evmod.Event(
                magnitudes=magnitudes,
                origins=[origin],
                resource_id=id_,
                event_descriptions=descs
            )
            events.append(event)
            if format_check:
                return True
    if format_check:
        # empty file will return an empty catalog,
        # but it is not detected as CSV file
        return False
    return evmod.Catalog(events=events)


def _write_csv(events, fname, fields='basic', depth_in_km=True, delimiter=',',
               header=None):
    """
    Write ObsPy catalog to CSV file

    :param events: catalog or list of events
    :param fname: filename or file-like object to write to
    :param fields: set format and header names of CSV file, see
       :const:`obspy.io.csv.FIELDS`, you can use your own format string here,
       just make sure to use the pre-defined header names
    :param depth_in_km: write depth in units of kilometer (default: True) or
        meter
    :param delimiter: defaults to `','`, if the delimiter is changed, ObsPy's
        read_events function will not automatically identify the file as
        CSV file
    :param header: Use a non-default header row

    .. warning::
        If the parameters `delimiter` or `header` are changed,
        ObsPy's read_events function will not automatically identify the file
        as CSV file

    Example
    -------

    >>> from obspy import read_events
    >>> events = read_events()  # get example catalog
    >>> events.write('local_catalog.csv', 'CSV')  # declare 'CSV' as format
    >>> with open('local_catalog.csv') as f: print(f.read())
    id,time,lat,lon,dep,magtype,mag
    20120404_0000041,2012-04-04T14:21:42.30000,41.818000,79.689000,1.000,mb,4.40
    20120404_0000038,2012-04-04T14:18:37.00000,39.342000,41.044000,14.400,ML,4.30
    20120404_0000039,2012-04-04T14:08:46.00000,38.017000,37.736000,7.000,ML,3.00
    <BLANKLINE>
    """
    fields = FIELDS.get(fields, fields)
    if ' ' in fields:
        fields = fields.split()
    fmtstr = delimiter.join(fields)
    if not depth_in_km and 'depm' not in fmtstr:
        fmtstr = fmtstr.replace('dep', 'depm')
    if header is None:
        fieldnames = [
            fn for _, fn, _, _ in Formatter().parse(fmtstr) if fn is not None]
        header = delimiter.join(fieldnames)
    with _open(fname, 'w') as f:
        f.write(header + '\n')
        for event in events:
            evid = str(event.resource_id).split('/')[-1]
            try:
                origin = _origin(event)
            except Exception:
                warn(f'No origin found -> do not write event {evid}')
                continue
            try:
                author = origin.creation_info.author
            except AttributeError:
                author = ''
            try:
                contrib = origin.creation_info.agency_id
            except AttributeError:
                contrib = ''
            try:
                magnitude = event.preferred_magnitude() or event.magnitudes[0]
            except Exception:
                warn(f'No magnitude found for event {evid}')
                mag = float('nan')
                magtype = ''
                magauthor = ''
            else:
                mag = magnitude.mag
                magtype = magnitude.magnitude_type or ''
                try:
                    magauthor = magnitude.creation_info.author
                except AttributeError:
                    magauthor = ''
            try:
                dep = origin.depth / (1000 if depth_in_km else 1)
            except Exception:
                warn(f'No depth set for event {evid}')
                dep = float('nan')
            for description in event.event_descriptions:
                if str(description.type) == 'region name':
                    region = description.text
                    break
            else:
                region = ''
            d = {'time': origin.time,
                 'lat': origin.latitude,
                 'lon': origin.longitude,
                 'dep' if depth_in_km else 'depm': dep,
                 'mag': mag,
                 'magtype': magtype,
                 'id': evid,
                 'author': author,
                 'contrib': contrib,
                 'magauthor': magauthor,
                 'region': region,
                 }
            f.write(fmtstr.format(**d).replace('nan', '') + '\n')


@map_example_filename('fname')
def load_csv(fname, skipheader=0, only=None, names=None,
             delimiter=',', **kw):
    """
    Load CSV or CSZ file into numpy array

    :param skipheader, names: see :func:`_read_csv`
    :param only: sequence, read only columns speified by name
    :param **kw: Other kwargs are passed to :func:`numpy.loadtxt`

    For an example see :mod:`obspy.io.csv`.
    """
    if isinstance(fname, str) and zipfile.is_zipfile(fname):
        with zipfile.ZipFile(fname) as zipf:
            with io.TextIOWrapper(
                    zipf.open('events.csv'), encoding='utf-8') as f:
                return load_csv(f)
    with _open(fname) as f:
        for _ in range(skipheader):
            f.readline()
        if names is None:
            names = f.readline().strip().split(',')
        names = _names_sequence(names)
        dtype = [(n, DTYPE[n]) for n in names if n in DTYPE and
                 (only is None or n in only)]
        usecols = [i for i, n in enumerate(names) if n in DTYPE and
                   (only is None or n in only)]
        kw.setdefault('usecols', usecols)
        kw.setdefault('dtype', dtype)
        return np.genfromtxt(f, delimiter=delimiter, **kw)


def _events2array(events, **kw):
    """
    Convert ObsPy catalog to numpy array

    All kwargs are passed to :func:`load_csv`, e.g. use
    `only=('lat', 'lon', 'mag')`
    to get an array with lat, lon, mag parameters.

    Example
    -------

    >>> from obspy import read_events
    >>> from obspy.io.csv import _events2array
    >>> events = read_events()
    >>> t = _events2array(events)
    >>> print(t.dtype.names)
    ('id', 'time', 'lat', 'lon', 'dep', 'magtype', 'mag')
    >>> print(t)   # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    [ ('20120404_0000041', '2012-04-04...', 41.818, 79.689, 1. , 'mb', 4.4)
     ('20120404_0000038', '2012-04-04T...', 39.342, 41.044, 14.4, 'ML', 4.3)
     ('20120404_0000039', '2012-04-04T...', 38.017, 37.736, 7. , 'ML', 3. )]
    >>> print(t['mag'])
    [ 4.4  4.3  3. ]
    """
    with io.StringIO() as f:
        _write_csv(events, f)
        f.seek(0)
        return load_csv(f, **kw)


def _read_eventtxt(fname, default=None, format_check=False):
    """
    Read EVENTTXT file and return ObsPy catalog

    :param fname: filename or file-like object to read from
    :param default: dictionary with default values, at the moment only
         magtype is supported,
         i.e. to set magtypes use ``default={'magtype': 'Ml'}``
    :param format_check: Check if the first event can be read

    For an example see :mod:`obspy.io.csv`.
    """
    return _read_csv(fname,
                     skipheader=1, names=EVENTTXT_NAMES, delimiter='|',
                     default=default, format_check=format_check)


def _write_eventtxt(events, fname):
    """
    Write ObsPy catalog to EVENTTXT file

    :param events: catalog or list of events
    :param fname: filename or file-like object to write to

    For an example see :mod:`obspy.io.csv`.
    """
    return _write_csv(events, fname, fields='eventtxt', header=EVENTTXT_HEADER,
                      delimiter='|')


@map_example_filename('fname')
def load_eventtxt(fname, **kw):
    """
    Load EVENTTXT file into numpy array

    For possible arguments see :func:`load_csv`.
    For an example see :mod:`obspy.io.csv`.
    """
    kw.setdefault('delimiter', '|')
    kw.setdefault('skipheader', 1)
    return load_csv(fname, names=EVENTTXT_NAMES, **kw)


def _read_picks(event, fname):
    """
    Read picks from CSV file and add them to event
    """
    otime = _origin(event).time
    picks = []
    arrivals = []
    with _open(fname) as f:
        reader = csv.DictReader(f)
        for row in reader:
            phase = row['phase']
            seedid = row['seedid']
            wid = (evmod.WaveformStreamID(seed_string=seedid) if seedid
                   else None)
            pick = evmod.Pick(waveform_id=wid, phase_hint=phase,
                              time=otime + float(row['time']))
            arrival = evmod.Arrival(phase=phase, pick_id=pick.resource_id,
                                    time_weight=float(row['weight']))
            picks.append(pick)
            arrivals.append(arrival)
    event.picks = picks
    event.origins[0].arrivals = arrivals


def _write_picks(event, fname, fields_picks='basic', delimiter=','):
    """
    Write picks from event to a CSV file
    """
    fields = PFIELDS.get(fields_picks, fields_picks)
    if ' ' in fields:
        fields = fields.split()
    fmtstr = delimiter.join(fields)
    fieldnames = [
        fn for _, fn, _, _ in Formatter().parse(fmtstr) if fn is not None]
    origin = _origin(event)
    weights = {str(arrival.pick_id): arrival.time_weight
               for arrival in origin.arrivals if arrival.time_weight}
    phases = {str(arrival.pick_id): arrival.phase
              for arrival in origin.arrivals if arrival.phase}
    with _open(fname, 'w') as f:
        f.write(delimiter.join(fieldnames) + '\n')
        for pick in event.picks:
            pick_id = str(pick.resource_id)
            try:
                seedid = pick.waveform_id.id
                if seedid is None:
                    raise AttributeError
            except AttributeError:
                warn(f'No waveform id found for pick {pick_id}')
                seedid = ''
            d = {'time': pick.time - origin.time,
                 'seedid': seedid,
                 'phase': phases.get(pick_id, pick.phase_hint),
                 'weight': weights.get(pick_id, 1.)}
            f.write(fmtstr.format(**d) + '\n')


def _read_csz(fname, default=None):
    """
    Read a CSZ file and return ObsPy catalog with picks

    :param fname: filename or file-like object to read from
    :param default: dictionary with default values, at the moment only
         magtype is supported,
         i.e. to set magtypes use ``default={'magtype': 'Ml'}``

    For an example see :mod:`obspy.io.csv`.
    """
    with zipfile.ZipFile(fname) as zipf:
        with io.TextIOWrapper(zipf.open('events.csv'), encoding='utf-8') as f:
            events = _read_csv(f, default=default)
        for event in events:
            evid = _evid(event)
            fname = f'picks_{evid}.csv'
            if fname not in zipf.namelist():
                continue
            with io.TextIOWrapper(zipf.open(fname), encoding='utf-8') as f:
                _read_picks(event, f)
    return events


def _write_csz(events, fname, fields='basic', fields_picks='basic', **kwargs):
    """
    Write ObsPy catalog to CSZ file

    :param events: catalog or list of events
    :param fname: filename or file-like object to write to
    :param fields: set format and header names of CSV file, see
       :const:`obspy.io.csv.FIELDS`, you can use your own format string here,
       just make sure to use the pre-defined header names
    :param fields_picks: set format and header names of CSV pick files, see
       :const:`obspy.io.csv.PFIELDS`, you can use your own format string here,
       just make sure to use the pre-defined header names
    :param **kwargs:
 compression and compression level can be specified see
 `zipfile doc <https://docs.python.org/library/zipfile.html#zipfile.ZipFile>`_:
        ```
        events.write('CSZ', compression=True, compresslevel=9)
        ```

    For an example see :mod:`obspy.io.csv`.
    """
    # allow True as value for compression
    if kwargs.get('compression') is True:
        kwargs['compression'] = zipfile.ZIP_DEFLATED
    with zipfile.ZipFile(fname, mode='w', **kwargs) as zipf:
        zipf.comment = CSZ_COMMENT
        with io.StringIO() as f:
            _write_csv(events, f, fields=fields)
            zipf.writestr('events.csv', f.getvalue())
        for event in events:
            if len(event.picks) == 0:
                continue
            evid = str(event.resource_id).split('/')[-1]
            try:
                _origin(event)
            except Exception:
                continue
            with io.StringIO() as f:
                _write_picks(event, f, fields_picks=fields_picks)
                zipf.writestr(f'picks_{evid}.csv', f.getvalue())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The libgse1 test suite.
"""
from obspy.io.gse2 import libgse1
from obspy.io.gse2.libgse2 import ChksumError
import pytest


class TestLibGSE1():
    """
    Test cases for libgse1.
    """
    def test_verify_checksums(self, testdata):
        """
        Tests verifying checksums for CM6 encoded GSE1 files.
        """
        # 1
        fh = open(testdata['acc.gse'], 'rb')
        libgse1.read(fh, verify_chksum=True)
        fh.close()
        # 2
        fh = open(testdata['y2000.gse'], 'rb')
        libgse1.read(fh, verify_chksum=True)
        fh.close()
        # 3
        fh = open(testdata['loc_STAU20031119011659.z'], 'rb')
        libgse1.read(fh, verify_chksum=True)
        fh.close()
        # 4 - second checksum is wrong
        fh = open(testdata['GRF_031102_0225.GSE.wrong_chksum'], 'rb')
        libgse1.read(fh, verify_chksum=True)  # correct
        with pytest.raises(ChksumError):
            libgse1.read(fh, verify_chksum=True)
        fh.close()



# -*- coding: utf-8 -*-
"""
The obspy.imaging.backend test suite.
"""
import matplotlib


class TestBackend:
    """
    Test cases for matplotlib backend.

    Note: This test will fail when called from an interactive Python session
    where matplotlib was already imported.
    """
    def test_backend(self):
        """
        Test to see if tests are running without any X11 or any other display
        variable set. Therefore, the AGG backend is chosen in
        obspy.imaging.tests.__init__, and nothing must be imported before,
        e.g. by obspy.imaging.__init__. The AGG backend does not require and
        display setting. It is therefore the optimal for programs on servers
        etc.
        """
        assert 'AGG' == matplotlib.get_backend().upper()



import numpy as np
import matplotlib.pyplot as plt

from obspy.signal.invsim import paz_to_freq_resp


poles = [-4.440 + 4.440j, -4.440 - 4.440j, -1.083 + 0.0j]
zeros = [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j]
scale_fac = 0.4

h, f = paz_to_freq_resp(poles, zeros, scale_fac, 0.005, 16384, freq=True)

plt.figure()
plt.subplot(121)
plt.loglog(f, abs(h))
plt.xlabel('Frequency [Hz]')
plt.ylabel('Amplitude')

plt.subplot(122)
phase = 2 * np.pi + np.unwrap(np.angle(h))
plt.semilogx(f, phase)
plt.xlabel('Frequency [Hz]')
plt.ylabel('Phase [radian]')
# ticks and tick labels at multiples of pi
plt.yticks(
    [0, np.pi / 2, np.pi, 3 * np.pi / 2, 2 * np.pi],
    ['$0$', r'$\frac{\pi}{2}$', r'$\pi$', r'$\frac{3\pi}{2}$', r'$2\pi$'])
plt.ylim(-0.2, 2 * np.pi + 0.2)
# title, centered above both subplots
plt.suptitle('Frequency Response of LE-3D/1s Seismometer')
# make more room in between subplots for the ylabel of right plot
plt.subplots_adjust(wspace=0.3)
plt.show()



# -*- coding: utf-8 -*-
"""
Various additional utilities for ObsPy xseed.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import warnings

from obspy import UTCDateTime


# Ignore Attributes of Blockettes
IGNORE_ATTR = ['blockette_id', 'blockette_name', 'compact', 'debug',
               'seed_version', 'strict', 'xseed_version',
               'length_of_blockette', 'blockette_type']


class SEEDParserException(Exception):
    pass


class SEEDAbbreviationNotFoundWarning(UserWarning):
    pass


def to_tag(name):
    """
    Creates a XML tag from a given string.
    """
    temp = name.lower().replace(' ', '_')
    temp = temp.replace('fir_', 'FIR_')
    temp = temp.replace('a0_', 'A0_')
    return temp


def to_string(tag):
    """
    Creates a pretty string from any given XML tag.
    """
    temp = tag.replace('_', ' ').title()
    temp = temp.replace('Fir', 'FIR')
    return temp


def datetime_2_string(dt, compact=False):
    """
    Generates a valid SEED time string from a UTCDateTime object.
    """
    if isinstance(dt, UTCDateTime):
        return dt.format_seed(compact)
    elif isinstance(dt, str):
        dt = dt.strip()
    if not dt:
        return ""
    try:
        dt = UTCDateTime(dt)
        return dt.format_seed(compact)
    except Exception:
        raise Exception("Invalid datetime %s: %s" % (type(dt), str(dt)))


def compare_seed(seed1, seed2):
    """
    Compares two SEED files.

    Only works with a record length of 4096 bytes.
    """
    # Each SEED string should be a multiple of the record length.
    if (len(seed1) % 4096) != 0:
        msg = "Length of first SEED string should be a multiple of 4096 bytes"
        raise Exception(msg)
    if (len(seed2) % 4096) != 0:
        msg = "Length of second SEED string should be a multiple of 4096 bytes"
        raise Exception(msg)
    # Loop over each record and remove empty ones. obspy.io.xseed doesn't write
    # empty records. Redundant code to ease coding...
    recnums = len(seed1) // 4096
    new_seed1 = b''
    for _i in range(recnums):
        cur_record = seed1[_i * 4096 + 8:(_i + 1) * 4096].strip()
        if cur_record == b'':
            continue
        new_seed1 += seed1[_i * 4096:(_i + 1) * 4096]
    seed1 = new_seed1
    recnums = len(seed2) // 4096
    new_seed2 = b''
    for _i in range(recnums):
        cur_record = seed2[_i * 4096 + 8:(_i + 1) * 4096].strip()
        if cur_record == b'':
            continue
        new_seed2 += seed2[_i * 4096:(_i + 1) * 4096]
    seed2 = new_seed2
    # length should be the same
    if len(seed1) != len(seed2):
        msg = "Length of SEED strings differ! (%d != %d)" % (len(seed1),
                                                             len(seed2))
        raise Exception(msg)
    # version string is always ' 2.4' for output
    if seed1[15:19] == b' 2.3':
        seed1 = seed1.replace(b' 2.3', b' 2.4', 1)
    if seed1[15:19] == b'02.3':
        seed1 = seed1.replace(b'02.3', b' 2.4', 1)
    # check for missing '~' in blockette 10 (faulty dataless from BW network)
    l = int(seed1[11:15])   # NOQA
    temp = seed1[0:(l + 8)]
    if temp.count(b'~') == 4:
        # added a '~' and remove a space before the next record
        # record length for now 4096
        b_l = ('%04i' % (l + 1)).encode('ascii', 'strict')
        seed1 = seed1[0:11] + b_l + seed1[15:(l + 8)] + b'~' + \
            seed1[(l + 8):4095] + seed1[4096:]
    # check each byte
    for i in range(0, len(seed1)):
        if seed1[i] == seed2[i]:
            continue
        temp = seed1[i] + seed2[i]
        if temp == b'0+':
            continue
        if temp == b'0 ':
            continue
        if temp == b' +':
            continue
        if temp == b'- ':
            # -056.996398+0031.0
            #  -56.996398  +31.0
            continue


def lookup_code(blockettes, blkt_number, field_name, lookup_code,
                lookup_code_number):
    """
    Loops over a list of blockettes until it finds the blockette with the
    right number and lookup code.
    """
    # List of all possible names for lookup
    for blockette in blockettes:
        if blockette.id != blkt_number:
            continue
        if getattr(blockette, lookup_code) != lookup_code_number:
            continue
        return getattr(blockette, field_name)
    return None


def format_resp(number, digits=4):
    """
    Formats a number according to the RESP format.
    """
    format_string = "%%-10.%dE" % digits
    return format_string % (number)


def blockette_34_lookup(abbr, lookup):
    """
    Gets certain values from blockette 34. Needed for RESP output.
    """
    try:
        l1 = lookup_code(abbr, 34, 'unit_name', 'unit_lookup_code', lookup)
        l2 = lookup_code(abbr, 34, 'unit_description', 'unit_lookup_code',
                         lookup)
        return l1 + ' - ' + l2
    except Exception:
        msg = 'Warning: Abbreviation reference for "%i" not found.' % lookup
        warnings.warn(msg, SEEDAbbreviationNotFoundWarning)
        return 'No Abbreviation Referenced'


def set_xpath(blockette, identifier):
    """
    Returns an X-Path String to a blockette with the correct identifier.
    """
    try:
        identifier = int(identifier)
    except Exception:
        msg = 'X-Path identifier needs to be an integer.'
        raise TypeError(msg)
    abbr_path = '/xseed/abbreviation_dictionary_control_header/'
    end_of_path = '[text()="%s"]/parent::*'
    if blockette == 30:
        return abbr_path + \
            'data_format_dictionary/data_format_identifier_code' + \
            end_of_path % identifier
    elif blockette == 31:
        return abbr_path + \
            'comment_description/comment_code_key' + \
            end_of_path % identifier
    elif blockette == 33:
        return abbr_path + \
            'generic_abbreviation/abbreviation_lookup_code' + \
            end_of_path % identifier
    elif blockette == 34:
        return abbr_path + \
            'units_abbreviations/unit_lookup_code' + \
            end_of_path % identifier
    # All dictionary blockettes.
    elif blockette == 'dictionary':
        return abbr_path + \
            '*/response_lookup_key' + \
            end_of_path % identifier
    msg = 'XPath for blockette %d not implemented yet.' % blockette
    raise NotImplementedError(msg)


def get_xpath(xpath):
    """
    Returns lookup key of XPath expression on abbreviation dictionary.
    """
    return int(xpath.split('"')[-2])


def unique_list(seq):
    # Not order preserving
    keys = {}
    for e in seq:
        keys[e] = 1
    return list(keys.keys())



# -*- coding: utf-8 -*-
"""
Various additional utilities for ObsPy.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import contextlib
import importlib.metadata
import inspect
import io
import itertools
import math
import os
import shutil
import sys
import tempfile
import warnings
from subprocess import STDOUT, CalledProcessError, check_output


import numpy as np

WIN32 = sys.platform.startswith('win32')

# The following dictionary maps the first character of the channel_id to the
# lowest sampling rate this so called Band Code should be used for according
# to SEED MANUAL p.124
# It can e.g. be used to request additional samples on both start and end to
# cut to the samples that really are nearest to requested start/end time
# afterwards.
BAND_CODE = {'F': 1000.0,
             'G': 1000.0,
             'D': 250.0,
             'C': 250.0,
             'E': 80.0,
             'S': 10.0,
             'H': 80.0,
             'B': 10.0,
             'M': 1.0,
             'L': 1.0,
             'V': 0.1,
             'U': 0.01,
             'R': 0.0001,
             'P': 0.000001,
             'T': 0.0000001,
             'Q': 0.00000001
             }

# Dict that stores results from load entry points
_ENTRY_POINT_CACHE = {}

# The kwargs used by load_entry_point function
_LOAD_ENTRY_POINT_KEYS = ('dist', 'group', 'name')


def guess_delta(channel):
    """
    Estimate time delta in seconds between each sample from given channel name.

    :type channel: str
    :param channel: Channel name, e.g. ``'BHZ'`` or ``'H'``
    :rtype: float
    :return: Returns ``0`` if band code is not given or unknown.

    .. rubric:: Example

    >>> print(guess_delta('BHZ'))
    0.1

    >>> print(guess_delta('H'))
    0.0125

    >>> print(guess_delta('XZY'))  # doctest: +SKIP
    0
    """
    try:
        return 1. / BAND_CODE[channel[0]]
    except Exception:
        msg = "No or unknown channel id provided. Specifying a channel id " + \
              "could lead to better selection of first/last samples of " + \
              "fetched traces."
        warnings.warn(msg)
    return 0


def score_at_percentile(values, per, limit=(), issorted=True):
    """
    Calculates the score at the given per percentile of the sequence a.

    For example, the score at ``per=50`` is the median.

    If the desired quantile lies between two data points, we interpolate
    between them.

    If the parameter ``limit`` is provided, it should be a tuple (lower,
    upper) of two values.  Values of ``a`` outside this (closed) interval
    will be ignored.

    .. rubric:: Examples

    >>> a = [1, 2, 3, 4]
    >>> score_at_percentile(a, 25)
    1.75
    >>> score_at_percentile(a, 50)
    2.5
    >>> score_at_percentile(a, 75)
    3.25

    >>> a = [6, 47, 49, 15, 42, 41, 7, 255, 39, 43, 40, 36, 500]
    >>> score_at_percentile(a, 25, limit=(0, 100))
    25.5
    >>> score_at_percentile(a, 50, limit=(0, 100))
    40
    >>> score_at_percentile(a, 75, limit=(0, 100))
    42.5

    This function is taken from `scipy.stats.score_at_percentile`.

    Copyright (c) Gary Strangman
    """
    if limit:
        values = [v for v in values if limit[0] < v < limit[1]]

    if issorted:
        values = sorted(values)

    def _interpolate(a, b, fraction):
        return a + (b - a) * fraction

    idx = per / 100. * (len(values) - 1)
    if (idx % 1 == 0):
        return values[int(idx)]
    else:
        return _interpolate(values[int(idx)], values[int(idx) + 1], idx % 1)


def flat_not_masked_contiguous(a):
    """
    Find contiguous unmasked data in a masked array along the given axis.

    This function is taken from
    :func:`numpy.ma.flatnotmasked_contiguous`.

    Copyright (c) Pierre Gerard-Marchant
    """
    np.ma.extras.flatnotmasked_contiguous
    m = np.ma.getmask(a)
    if m is np.ma.nomask:
        return slice(0, a.size, None)
    i = 0
    result = []
    for (k, g) in itertools.groupby(m.ravel()):
        n = len(list(g))
        if not k:
            result.append(slice(i, i + n))
        i += n
    return result


def complexify_string(line):
    """
    Converts a string in the form "(real, imag)" into a complex type.

    :type line: str
    :param line: String in the form ``"(real, imag)"``.
    :rtype: complex
    :return: Complex number.

    .. rubric:: Example

    >>> complexify_string("(1,2)")
    (1+2j)

    >>> complexify_string(" ( 1 , 2 ) ")
    (1+2j)
    """
    temp = line.split(',')
    return complex(float(temp[0].strip()[1:]), float(temp[1].strip()[:-1]))


def to_int_or_zero(value):
    """
    Converts given value to an integer or returns 0 if it fails.

    :param value: Arbitrary data type.
    :rtype: int

    .. rubric:: Example

    >>> to_int_or_zero("12")
    12

    >>> to_int_or_zero("x")
    0
    """
    try:
        return int(value)
    except ValueError:
        return 0


def get_untracked_files_from_git():
    """
    Tries to return a list of files (absolute paths) that are untracked by git
    in the repository.

    Returns `None` if the system call to git fails.
    """
    dir_ = os.path.abspath(
        os.path.dirname(inspect.getfile(inspect.currentframe())))
    dir_ = os.path.dirname(os.path.dirname(os.path.dirname(dir_)))
    try:
        # Check that the git root directory is actually the ObsPy directory.
        p = check_output(['git', 'rev-parse', '--show-toplevel'],
                         cwd=dir_, stderr=STDOUT)
        git_root_dir = p.decode().strip()
        if git_root_dir:
            git_root_dir = os.path.abspath(git_root_dir)
        if os.path.normcase(git_root_dir) != os.path.normcase(dir_):
            raise ValueError('Git root directory (%s) does not match expected '
                             'path (%s).' % (git_root_dir, dir_))
        p = check_output(['git', 'status', '-u', '--porcelain'],
                         cwd=dir_, stderr=STDOUT)
        stdout = p.decode().splitlines()
        files = [os.path.abspath(os.path.join(dir_, line.split()[1].strip()))
                 for line in stdout
                 if line.startswith("??")]
    except (OSError, CalledProcessError):
        return None
    return files


class CaptureIO(io.TextIOWrapper):
    def __init__(self):
        super(CaptureIO, self).__init__(io.BytesIO(), encoding='utf-8',
                                        newline='\n', write_through=True)

    def getvalue(self):
        return self.buffer.getvalue().decode('utf-8')


@contextlib.contextmanager
def CatchOutput():  # NOQA
    """
    A context manager that captures input to stdout/stderr. Python level only!

    Always use with "with" statement. Does nothing otherwise.

    >>> with CatchOutput() as out:  # doctest: +SKIP
    ...    sys.stdout.write("mystdout")
    ...    sys.stderr.write("mystderr")
    >>> print(out.stdout)  # doctest: +SKIP
    mystdout
    >>> print(out.stderr)  # doctest: +SKIP
    mystderr
    """
    # Dummy class to transport the output.
    class Output():
        stdout = ''
        stderr = ''
    out = Output()

    # set current stdout/stderr to in-memory text streams
    sys.stdout = stdout_result = CaptureIO()
    sys.stderr = stderr_result = CaptureIO()

    try:
        raised = False
        yield out
    except SystemExit:
        raised = True
    finally:
        out.stdout = stdout_result.getvalue()
        out.stderr = stderr_result.getvalue()

        # reset to original stdout/stderr
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

        # normalize line breaks
        if WIN32:
            out.stdout = out.stdout.replace('\r', '')
            out.stderr = out.stderr.replace('\r', '')

    if raised:
        raise SystemExit(out.stderr)


def _py3_windowsconsoleio_workaround():
    """
    This monkey patch prevents crashing Py >=3.6 under Windows while using
    the SuppressOutput context manager.

    Python 3.6 implemented unicode console handling for Windows. This works
    by reading/writing to the raw console handle using
    ``{Read,Write}ConsoleW``.
    The problem is that we are going to ``dup2`` over the stdio file
    descriptors when doing ``FDCapture`` and this will ``CloseHandle`` the
    handles used by Python to write to the console. Though there is still some
    weirdness and the console handle seems to only be closed randomly and not
    on the first call to ``CloseHandle``, or maybe it gets reopened with the
    same handle value when we suspend capturing.
    The workaround in this case will reopen stdio with a different fd which
    also means a different handle by replicating the logic in
    "Py_lifecycle.c:initstdio/create_stdio".

    See https://github.com/pytest-dev/py/issues/103
    See http://bugs.python.org/issue30555
    See https://github.com/obspy/obspy/issues/3148#issuecomment-1274254892
    """
    if not WIN32 or sys.version_info[:2] < (3, 6):
        return
    if not hasattr(sys.stdout, 'buffer'):
        return
    buffered = hasattr(sys.stdout.buffer, 'raw')
    raw_stdout = sys.stdout.buffer.raw if buffered else sys.stdout.buffer

    if not isinstance(raw_stdout, io._WindowsConsoleIO):
        return

    def _reopen_stdio(f, mode):
        if not buffered and mode[0] == 'w':
            buffering = 0
        else:
            buffering = -1

        return io.TextIOWrapper(
            open(os.dup(f.fileno()), mode, buffering),
            f.encoding,
            f.errors,
            f.newlines,
            f.line_buffering)

    sys.stdin = _reopen_stdio(sys.stdin, 'rb')
    sys.stdout = _reopen_stdio(sys.stdout, 'wb')
    sys.stderr = _reopen_stdio(sys.stderr, 'wb')


@contextlib.contextmanager
def SuppressOutput():  # noqa
    """
    A context manager that suppresses output to stdout/stderr.
    Always use with "with" statement. Does nothing otherwise.
    >>> with SuppressOutput():  # doctest: +SKIP
    ...    os.system('echo "mystdout"')
    ...    os.system('echo "mystderr" >&2')

    Note: Does not work reliably for Python 3 under Windows - see
    function definition of _py3_windowsconsoleio_workaround().
    """
    _py3_windowsconsoleio_workaround()
    with os.fdopen(os.dup(1), 'wb', 0) as tmp_stdout:
        with os.fdopen(os.dup(2), 'wb', 0) as tmp_stderr:
            with open(os.devnull, 'wb') as to_file:
                sys.stdout.flush()
                sys.stderr.flush()
                os.dup2(to_file.fileno(), 1)
                os.dup2(to_file.fileno(), 2)
                try:
                    yield
                finally:
                    sys.stdout.flush()
                    sys.stderr.flush()
                    os.dup2(tmp_stdout.fileno(), 1)
                    os.dup2(tmp_stderr.fileno(), 2)
    # reset to original stdout/stderr
    sys.stdin = sys.__stdin__
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__


@contextlib.contextmanager
def TemporaryWorkingDirectory():  # noqa --> this name is IMHO ok for a CM
    """
    A context manager that changes to a temporary working directory.

    Always use with "with" statement. Does nothing useful otherwise.

    >>> with TemporaryWorkingDirectory():  # doctest: +SKIP
    ...    os.system('echo "$PWD"')
    """
    tempdir = tempfile.mkdtemp(prefix='obspy-')
    old_dir = os.getcwd()
    os.chdir(tempdir)
    try:
        yield tempdir
    finally:
        os.chdir(old_dir)
        try:
            shutil.rmtree(tempdir)
        # Windows on appveyor is having issues with removing these temporary
        # directories (e.g. https://ci.appveyor.com/project/obspy/obspy/builds
        # /20718605/job/0vuja1b95rv5a0s2).
        except Exception as e:
            warnings.warn(e.__repr__())


def factorize_int(x):
    """
    Calculate prime factorization of integer.

    Could be done faster but faster algorithm have much more lines of code and
    this is fast enough for our purposes.

    https://stackoverflow.com/q/14550794

    >>> factorize_int(1800004)
    [2, 2, 450001]
    >>> factorize_int(1800003)
    [3, 19, 23, 1373]
    """
    if x == 1:
        return [1]
    factors, limit, check, num = [], int(math.sqrt(x)) + 1, 2, x
    for check in range(2, limit):
        while num % check == 0:
            factors.append(check)
            num /= check
    if num > 1:
        factors.append(int(num))
    return factors


def get_window_times(starttime, endtime, window_length, step, offset,
                     include_partial_windows):
    """
    Function calculating a list of times making up equal length windows from
    within a given time interval.

    :param starttime: The start time of the whole time interval.
    :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param endtime: The end time of the whole time interval.
    :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param window_length: The length of each window in seconds.
    :type window_length: float
    :param step: The step between the start times of two successive
        windows in seconds. Can be negative if an offset is given.
    :type step: float
    :param offset: The offset of the first window in seconds relative to
        the start time of the whole interval.
    :type offset: float
    :param include_partial_windows: Determines if windows that are
        shorter then 99.9 % of the desired length are returned.
    :type include_partial_windows: bool
    """
    if step > 0:
        end = endtime.timestamp - 0.001 * step
    else:
        # This minus is correct here due to the way the actual window times
        # are calculate later on.
        end = starttime.timestamp - 0.001 * abs(step)
    # Left sides of each window.
    indices = np.arange(start=starttime.timestamp + offset,
                        stop=end, step=step, dtype=np.float64)
    if step > 0:
        # Generate all possible windows.
        windows = [(_i, min(_i + window_length, endtime.timestamp))
                   for _i in indices]
    else:
        # Generate all possible windows.
        windows = [(max(_i - window_length, starttime.timestamp), _i)
                   for _i in indices]

    # Potentially remove partial windows not fulfilling the window length
    # criterion.
    if not include_partial_windows:
        windows = [_i for _i in windows
                   if abs(_i[1] - _i[0]) > 0.999 * window_length]

    t = type(starttime)
    return [(t(_i[0]), t(_i[1])) for _i in windows]


class MatplotlibBackend(object):
    """
    A helper class for switching the matplotlib backend.

    Can be used as a context manager to temporarily switch the backend or by
    using the :meth:`~MatplotlibBackend.switch_backend` staticmethod.

    The context manager has no effect when setting ``backend=None``.

    :type backend: str
    :param backend: Name of matplotlib backend to switch to.
    :type sloppy: bool
    :param sloppy: If ``True``, uses :func:`matplotlib.pyplot.switch_backend`
        and no warning will be shown if the backend was not switched
        successfully. If ``False``, additionally tries to use
        :func:`matplotlib.use` first and also shows a warning if the backend
        was not switched successfully.
    :type close: bool
    :param close: Whether to close all matplotlib figures when exiting the
        context manager.
    """

    def __init__(self, backend, sloppy=True, close=False):
        self.temporary_backend = backend
        self.sloppy = sloppy
        import matplotlib
        self.previous_backend = matplotlib.get_backend()
        self.close = close

    def __enter__(self):
        if self.temporary_backend is None:
            return
        self.switch_backend(backend=self.temporary_backend, sloppy=self.sloppy)

    def __exit__(self, exc_type, exc_val, exc_tb):  # @UnusedVariable
        if self.close:
            import matplotlib.pyplot as plt
            plt.close('all')
        if self.temporary_backend is None:
            return
        self.switch_backend(backend=self.previous_backend, sloppy=self.sloppy)

    @staticmethod
    def switch_backend(backend, sloppy=True):
        """
        Switch matplotlib backend.

        :type backend: str
        :param backend: Name of matplotlib backend to switch to.
        :type sloppy: bool
        :param sloppy: If ``True``, only uses
            :func:`matplotlib.pyplot.switch_backend` and no warning will be
            shown if the backend was not switched successfully. If ``False``,
            additionally tries to use :func:`matplotlib.use` first and also
            shows a warning if the backend was not switched successfully.
        """
        import matplotlib
        # first of all, all figures should be closed, matplotlib is showing a
        # warning that this will not be done automatically by matplotlib in
        # newer releases anymore, so do it here
        import matplotlib.pyplot as plt
        plt.close("all")
        # sloppy. only do a `plt.switch_backend(..)`
        if sloppy:
            plt.switch_backend(backend)
        else:
            # check if `matplotlib.use(..)` is emitting a warning
            try:
                with warnings.catch_warnings(record=True):
                    warnings.simplefilter("error", UserWarning)
                    matplotlib.use(backend)
            # if that's the case, follow up with `plt.switch_backend(..)`
            except UserWarning:
                plt.switch_backend(backend)
            # finally check if the switch was successful,
            # show a warning if not
            if matplotlib.get_backend().upper() != backend.upper():
                msg = "Unable to change matplotlib backend to '%s'" % backend
                raise Exception(msg)


def limit_numpy_fft_cache(max_size_in_mb_per_cache=100):
    """
    NumPy's FFT implementation utilizes caches to speedup subsequent FFTs of
    the same size. This accumulates memory when run for various length FFTs
    as can readily happen in seismology.

    This utility function clears both, full and real-only caches if their
    size is above the given threshold.

    The default 100 MB is fairly generous but we still want to profit from
    the cache where applicable.
    """
    for cache in ["_fft_cache", "_real_fft_cache"]:
        # Guard against different numpy versions just to be safe.
        if not hasattr(np.fft.fftpack, cache):
            continue
        cache = getattr(np.fft.fftpack, cache)
        # Check type directly and don't use isinstance() as future numpy
        # versions might use some subclass or what not.
        if type(cache) is not dict:
            continue
        # Its a dictionary with list's of arrays as the values. Wrap in
        # try/except to guard against future numpy changes.
        try:
            total_size = sum([_j.nbytes for _i in cache.values() for _j in _i])
        except Exception:
            continue
        if total_size > max_size_in_mb_per_cache * 1024 * 1024:
            cache.clear()


def buffered_load_entry_point(dist, group, name):
    """
    Return `name` entry point of `group` for `dist` or raise ImportError
    :type dist: str
    :param dist: The name of the distribution containing the entry point.
    :type group: str
    :param group: The name of the group containing the entry point.
    :type name: str
    :param name: The name of the entry point.
    :return: The loaded entry point
    """
    hash_str = '/'.join([dist, group, name])
    if hash_str not in _ENTRY_POINT_CACHE:
        if sys.version_info.minor < 10:
            # compatibility workaround for Python 3.8 and 3.9
            # on these versions, entry points don't have their distribution
            # information attached to them, so we have to check if any matching
            # entry point we find is listed in that distribution's entry point
            # list
            dist_eps = (importlib.metadata.distribution(group.split(".")[0])
                        .entry_points)

            eps = list(
                ep for ep in
                dist_eps if f'{group}:{name}' == f"{ep.group}:{ep.name}"
            )
            if not len(eps):
                eps = list(
                    ep for ep in
                    dist_eps if f'{group}.{name}' == f"{ep.group}:{ep.name}"
                )
        else:
            eps = list(
                ep for ep in importlib.metadata.entry_points(
                    group=group, name=name)
                if ep.dist.name == dist)
        if len(set(eps)) > 1:
            warnings.warn(f'Multiple entry points matching:\n{eps!s}')
        elif not len(eps):
            # resemble the exception pkg_resources would have raised here
            msg = f"Entry point ('{dist}', '{group}', '{name}') not found"
            raise ImportError(msg)
        ep = eps[0]
        _ENTRY_POINT_CACHE[hash_str] = ep.load()
    return _ENTRY_POINT_CACHE[hash_str]


def _yield_obj_parent_attr(obj, cls=None, is_attr=None, has_attr=None):
    """
    Recurse an object, yield a tuple of object, parent, attr.

    Can be used, for example, to yield all ResourceIdentifier instances
    contained in any obspy.core.event class instances and attached instances,
    as well as the objects they are attached to (parents) and the attribute
    name in which they are stored (attr).

    :param obj:
        The object to recurse through attributes of lists, tuples, and other
        instances.
    :param cls:
        Only return instances of cls if not None, else return all instances.
    :param is_attr:
        Only return objects stored as attr_name, if None return all.
    :param has_attr:
        Only return objects that have attribute has_attr, if None return all.

    .. rubric:: General Usage

    Get a list of all resource_ids contained in an event, the objects they
    are attached to, and the attribute name on the parent object.

    >>> import obspy
    >>> from obspy.core.event import ResourceIdentifier
    >>> cat = obspy.read_events()
    >>> resource_tuple = list(_yield_obj_parent_attr(cat, ResourceIdentifier))
    """
    ids = set()  # id cache to avoid circular references

    def func(obj, attr=None, parent=None):
        id_tuple = (id(obj), id(parent))

        # If object/parent combo have not been yielded continue.
        if id_tuple not in ids:
            ids.add(id_tuple)
            # Check if this object is stored as the desired attribute.
            is_attribute = is_attr is None or attr == is_attr
            # Check if the object has the desired attribute.
            has_attribute = has_attr is None or hasattr(obj, has_attr)
            # Check if isinstance of desired class.
            is_instance = cls is None or isinstance(obj, cls)
            # Yield object, parent, and attr if desired conditions are met.
            if is_attribute and has_attribute and is_instance:
                yield (obj, parent, attr)
            # Iterate through basic built-in types.
            if isinstance(obj, (list, tuple)):
                for val in obj:
                    for out in func(val, attr=attr, parent=obj):
                        yield out
            elif isinstance(obj, dict):
                for item, val in obj.items():
                    for out in func(val, attr=item, parent=obj):
                        yield out
            # Iterate through non built-in object attributes.
            elif hasattr(obj, '__slots__'):
                for attr in obj.__slots__:
                    val = getattr(obj, attr)
                    for out in func(val, attr=attr, parent=obj):
                        yield out
            elif hasattr(obj, '__dict__'):
                for item, val in obj.__dict__.items():
                    for out in func(val, attr=item, parent=obj):
                        yield out

    return func(obj)


def _yield_resource_id_parent_attr(obj):
    """
    Specialized form of _yield_obj_parent_attr for getting ResourceIdentifiers.

    This function makes some assumptions because only resource_identifiers are
    being sought in order to improve efficiency.
    """
    from obspy.core.event import ResourceIdentifier

    ids = set()  # id cache to avoid circular references

    def func(obj, parent=None, attr=None):
        if obj is None or not (hasattr(obj, '__dict__') or
                               isinstance(obj, (list, tuple))):
            return  # stop iteration
        id_tuple = (id(obj), id(parent))
        if id_tuple not in ids:
            ids.add(id_tuple)
            # Yield object, parent, and attr if desired conditions are met
            if isinstance(obj, ResourceIdentifier):
                yield (obj, parent, attr)
            # Iterate through basic built-in types.
            elif isinstance(obj, (list, tuple)):
                for val in obj:
                    for out in func(val, obj, attr):
                        yield out
            # Iterate through non built-in object attributes.
            elif hasattr(obj, '__dict__'):
                for item, val in obj.__dict__.items():
                    for out in func(val, obj, item):
                        yield out

    return func(obj)


def ptp(a, *args, **kwargs):
    """
    Replacement for :meth:`numpy.ndarray.ptp()` and the corresponding method on
    `MaskedArray` objects which are being removed in numpy 2.0
    Basically just makes sure we call the correct replacement function numpy
    put in place for regular and masked arrays.

    :type a: :class:`numpy.ndarray` or :class:`numpy.ma.MaskedArray`
    """
    if isinstance(a, np.ma.MaskedArray):
        return np.ma.ptp(a, *args, **kwargs)
    return np.ptp(a, *args, **kwargs)


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



import io
import os.path
from tempfile import gettempdir

import numpy as np
from obspy import read_events
from obspy.core.util import NamedTemporaryFile
import obspy.io.csv.core as iocsv
import pytest
from obspy.core.util import get_example_file


def test_csv():
    events = read_events()
    with NamedTemporaryFile(suffix='.csv') as ft:
        events.write(ft.name, 'CSV')
        events2 = read_events(ft.name)
    assert len(events2) == len(events)
    assert events2[0].origins[0].time == events[0].origins[0].time
    with NamedTemporaryFile(suffix='.csv') as ft:
        events.write(ft.name, 'CSV', depth_in_km=True)
        events2 = read_events(ft.name)
    assert len(events2) == len(events)
    assert events2[0].origins[0].time == events[0].origins[0].time


def test_csv_reading_external_catalog():
    names = 'year mon day hour minu sec _ lat lon dep mag id'
    incomplete_names = 'year mon day hour minu sec _ lat lon dep'
    fname = '/path/to/external.csv'
    assert not iocsv._is_csv(fname)
    events = read_events(fname, 'CSV', skipheader=1, names=names)
    events2 = read_events(fname, 'CSV', skipheader=1, names=incomplete_names)
    assert len(events) == 1
    assert str(events[0].origins[0].time) == '2023-05-06T19:55:01.300000Z'
    assert len(events[0].magnitudes) == 1
    assert len(events2) == 1
    assert str(events2[0].origins[0].time) == '2023-05-06T19:55:01.300000Z'
    assert len(events2[0].magnitudes) == 0


def test_csv_incomplete_catalog():
    events = read_events()
    del events[0].magnitudes[0].magnitude_type
    events[1].magnitudes = []
    events[1].preferred_magnitude_id = None
    events[1].origins[0].depth = None
    events[2].origins = []
    events[2].preferred_origin_id = None
    with NamedTemporaryFile(suffix='.csv') as ft:
        with pytest.warns(Warning, match='.*event 2012'):
            events.write(ft.name, 'CSV')
        events2 = read_events(ft.name, 'CSV')
    assert len(events2) == 2
    assert events2[0].origins[0].time == events[0].origins[0].time
    assert events2[1].origins[0].depth is None
    assert events2[1].origins[0].time == events[1].origins[0].time

    fname = get_example_file('incomplete.csv')
    assert iocsv._is_csv(fname)
    events = read_events(fname)
    assert len(events) == 2
    assert events[0].origins[0].depth is None
    assert len(events[0].magnitudes) == 0
    assert events[1].origins[0].depth is None
    assert len(events[1].magnitudes) == 1
    assert events[1].magnitudes[0].mag == 10.0


def test_csv_custom_fmt():
    events = read_events()
    with NamedTemporaryFile(suffix='.csv') as ft:
        fname = ft.name
        events.write(fname, 'CSV', fields='{lat:.5f} {lon:.5f}')
        assert not iocsv._is_csv(fname)
        data = np.genfromtxt(fname, names=True, delimiter=',')
        assert len(data) == 3
        assert len(data[0]) == 2


def test_csv_empty():
    assert not iocsv._is_csv(b'')
    empty_cat = iocsv._read_csv(b'')
    assert len(empty_cat) == 0


def test_csz(check_compression=False):
    events = read_events('/path/to/example.pha')
    tempdir = gettempdir()
    fname = os.path.join(tempdir, 'obbspycsv_testfile.csz')
    with NamedTemporaryFile(suffix='.csz') as ft:
        fname = ft.name

        def _test_write_read(events, **kw):
            events.write(fname, 'CSZ', **kw)
            assert iocsv._is_csz(fname)
            events2 = read_events(fname, check_compression=check_compression)
            assert len(events2) == len(events)
            for ev1, ev2 in zip(events, events2):
                assert len(ev2.origins[0].arrivals) == \
                    len(ev1.origins[0].arrivals)
                assert len(ev2.picks) == \
                    len(ev1.picks)
        _test_write_read(events)
        _test_write_read(events, compression=False)
        try:
            import zlib  # noqa: F401
        except ImportError:
            pass
        else:
            _test_write_read(events, compression=True, compresslevel=6)
        # test with missing origin and waveformid
        events[1].origins = []
        events[0].picks[0].waveform_id = None
        with pytest.warns(Warning,
                          match='The object with identity|No.*found'):
            events.write(fname, 'CSZ')
        assert iocsv._is_csz(fname)
        events2 = read_events(fname, check_compression=check_compression)
        assert len(events2) == 1
        assert (len(events2[0].origins[0].arrivals) ==
                len(events[0].origins[0].arrivals))
        assert len(events2[0].picks) == len(events[0].picks)
        assert events2[0].picks[0].waveform_id is None


def test_csz_without_picks(check_compression=False):
    events = read_events()
    with NamedTemporaryFile(suffix='.csz') as ft:
        fname = ft.name
        events.write(fname, 'CSZ')
        assert iocsv._is_csz(fname)
        events2 = read_events(fname, check_compression=check_compression)
        assert events2[0]._format == 'CSZ'
        assert len(events2) == len(events)


def test_csz_without_check_compression_parameters():
    test_csz(check_compression=True)
    test_csz_without_picks(check_compression=True)


def test_load_csv():
    events = read_events()
    with NamedTemporaryFile(suffix='.csv') as ft:
        events.write(ft.name, 'CSV')
        t = iocsv.load_csv(ft.name)
    assert t['mag'][0] == 4.4
    with NamedTemporaryFile(suffix='.csz') as ft:
        events.write(ft.name, 'CSZ')
        t = iocsv.load_csv(ft.name)
    assert t['mag'][0] == 4.4
    t = iocsv._events2array(events)
    assert t['mag'][0] == 4.4


def test_load_csv_incomplete_catalog():
    events = read_events()
    del events[0].magnitudes[0].magnitude_type
    events[1].magnitudes = []
    events[1].preferred_magnitude_id = None
    events[2].origins = []
    events[2].preferred_origin_id = None
    with NamedTemporaryFile(suffix='.csv') as ft:
        with pytest.warns(Warning, match='No.*found'):
            events.write(ft.name, 'CSV')
        t = iocsv.load_csv(ft.name)
    assert len(t) == 2
    assert np.isnan(t['mag'][1])


def test_load_csv_some_cols():
    events = read_events()
    with NamedTemporaryFile(suffix='.csv') as ft:
        fields = '{lat:.6f} {lon:.6f} {mag:.2f}'
        events.write(ft.name, 'CSV', fields=fields)
        t = iocsv.load_csv(ft.name)
        t2 = iocsv.load_csv(ft.name, only=['mag'])
        t3 = iocsv.load_csv(ft.name, skipheader=1, names={2: 'mag'})
    assert t['mag'][0] == 4.4
    assert t2['mag'][0] == 4.4
    assert t3['mag'][0] == 4.4
    assert 'mag' in t2.dtype.names
    assert 'lat' not in t2.dtype.names
    assert 'mag' in t3.dtype.names
    assert 'lat' not in t3.dtype.names


def test_events2array():
    events = read_events()
    t = iocsv._events2array(events)
    assert t['mag'][0] == 4.4


def test_eventtxt():
    fname = get_example_file('events.txt')
    assert iocsv._is_eventtxt(fname)
    events = read_events(fname)
    arr = iocsv.load_eventtxt(fname)
    assert len(events) == 2
    assert str(events[0].origins[0].time) == '2012-04-11T08:38:37.000000Z'
    assert events[0].origins[0].creation_info.author == 'ISC'
    assert events[0].magnitudes[0].creation_info.author == 'GCMT'
    assert events[0].event_descriptions[0].text == 'SUMATRA'
    assert len(events[0].magnitudes) == 1
    assert events[1].origins[0].creation_info is None
    assert events[1].magnitudes[0].creation_info is None
    assert list(arr['mag']) == [8.6, 8.5]
    with open(fname) as f:
        eventtxt = f.read()
    with io.StringIO() as f:
        events.write(f, 'EVENTTXT')
        assert f.getvalue() == eventtxt



# -*- coding: utf-8 -*-
"""
Testing utilities for ObsPy.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import difflib
import doctest
import inspect
import io
import os
import re

import numpy as np
from lxml import etree

MODULE_TEST_SKIP_CHECKS = {}


def compare_xml_strings(doc1, doc2):
    """
    Simple helper function to compare two XML byte strings.

    :type doc1: bytes
    :type doc2: bytes
    """
    obj1 = etree.fromstring(doc1).getroottree()
    obj2 = etree.fromstring(doc2).getroottree()

    buf = io.BytesIO()
    obj1.write_c14n(buf)
    buf.seek(0, 0)
    str1 = buf.read().decode()
    str1 = [_i.strip() for _i in str1.splitlines()]

    buf = io.BytesIO()
    obj2.write_c14n(buf)
    buf.seek(0, 0)
    str2 = buf.read().decode()
    str2 = [_i.strip() for _i in str2.splitlines()]

    unified_diff = difflib.unified_diff(str1, str2)

    err_msg = "\n".join(unified_diff)
    if err_msg:  # pragma: no cover
        raise AssertionError("Strings are not equal.\n" + err_msg)


def remove_unique_ids(xml_string, remove_creation_time=False):
    """
    Removes unique ID parts of e.g. 'publicID="..."' attributes from xml
    strings.

    :type xml_string: str
    :param xml_string: xml string to process
    :type remove_creation_time: bool
    :param xml_string: controls whether to remove 'creationTime' tags or not.
    :rtype: str
    """
    prefixes = ["id", "publicID", "pickID", "originID", "preferredOriginID",
                "preferredMagnitudeID", "preferredFocalMechanismID",
                "referenceSystemID", "methodID", "earthModelID",
                "triggeringOriginID", "derivedOriginID", "momentMagnitudeID",
                "greensFunctionID", "filterID", "amplitudeID",
                "stationMagnitudeID", "earthModelID", "slownessMethodID",
                "pickReference", "amplitudeReference"]
    if remove_creation_time:
        prefixes.append("creationTime")
    for prefix in prefixes:
        xml_string = re.sub("%s='.*?'" % prefix, '%s=""' % prefix, xml_string)
        xml_string = re.sub('%s=".*?"' % prefix, '%s=""' % prefix, xml_string)
        xml_string = re.sub("<%s>.*?</%s>" % (prefix, prefix),
                            '<%s/>' % prefix, xml_string)
    return xml_string


def get_all_py_files():
    """
    Return a list with full absolute paths to all .py files in ObsPy file tree.

    :rtype: list[str]
    """
    util_dir = os.path.abspath(inspect.getfile(inspect.currentframe()))
    obspy_dir = os.path.dirname(os.path.dirname(os.path.dirname(util_dir)))
    py_files = set()
    # Walk the obspy directory structure
    for dirpath, _, filenames in os.walk(obspy_dir):
        py_files.update([os.path.abspath(os.path.join(dirpath, i)) for i in
                         filenames if i.endswith(".py")])
    return sorted(py_files)


def create_diverse_catalog():
    """
    Create a catalog with a single event that has many features.

    Uses most the event related classes.
    """

    # imports are here in order to avoid circular import issues
    import obspy.core.event as ev
    from obspy import UTCDateTime, Catalog
    # local dict for storing state
    state = dict(time=UTCDateTime('2016-05-04T12:00:01'))

    def _create_event():
        event = ev.Event(
            event_type='mining explosion',
            event_descriptions=[_get_event_description()],
            picks=[_create_pick()],
            origins=[_create_origins()],
            station_magnitudes=[_get_station_mag()],
            magnitudes=[_create_magnitudes()],
            amplitudes=[_get_amplitudes()],
            focal_mechanisms=[_get_focal_mechanisms()],
        )
        # set preferred origin, focal mechanism, magnitude
        preferred_objects = dict(
            origin=event.origins[-1].resource_id,
            focal_mechanism=event.focal_mechanisms[-1].resource_id,
            magnitude=event.magnitudes[-1].resource_id,
        )
        for item, value in preferred_objects.items():
            setattr(event, 'preferred_' + item + '_id', value)

        event.scope_resource_ids()
        return event

    def _create_pick():
        # setup some of the classes
        creation = ev.CreationInfo(
            agency='SwanCo',
            author='Indago',
            creation_time=UTCDateTime(),
            version='10.10',
            author_url=ev.ResourceIdentifier('smi:local/me.com'),
        )

        pick = ev.Pick(
            time=state['time'],
            comments=[ev.Comment(x) for x in 'BOB'],
            evaluation_mode='manual',
            evaluation_status='final',
            creation_info=creation,
            phase_hint='P',
            polarity='positive',
            onset='emergent',
            back_azimith_errors={"uncertainty": 10},
            slowness_method_id=ev.ResourceIdentifier('smi:local/slow'),
            backazimuth=122.1,
            horizontal_slowness=12,
            method_id=ev.ResourceIdentifier(),
            horizontal_slowness_errors={'uncertainty': 12},
            filter_id=ev.ResourceIdentifier(),
            waveform_id=ev.WaveformStreamID('UU', 'FOO', '--', 'HHZ'),
        )
        state['pick_id'] = pick.resource_id
        return pick

    def _create_origins():
        ori = ev.Origin(
            resource_id=ev.ResourceIdentifier('smi:local/First'),
            time=UTCDateTime('2016-05-04T12:00:00'),
            time_errors={'uncertainty': .01},
            longitude=-111.12525,
            longitude_errors={'uncertainty': .020},
            latitude=47.48589325,
            latitude_errors={'uncertainty': .021},
            depth=2.123,
            depth_errors={'uncertainty': 1.22},
            depth_type='from location',
            time_fixed=False,
            epicenter_fixed=False,
            reference_system_id=ev.ResourceIdentifier(),
            method_id=ev.ResourceIdentifier(),
            earth_model_id=ev.ResourceIdentifier(),
            arrivals=[_get_arrival()],
            composite_times=[_get_composite_times()],
            quality=_get_origin_quality(),
            origin_type='hypocenter',
            origin_uncertainty=_get_origin_uncertainty(),
            region='US',
            evaluation_mode='manual',
            evaluation_status='final',
        )
        state['origin_id'] = ori.resource_id
        return ori

    def _get_arrival():
        return ev.Arrival(
            resource_id=ev.ResourceIdentifier('smi:local/Ar1'),
            pick_id=state['pick_id'],
            phase='P',
            time_correction=.2,
            azimuth=12,
            distance=10,
            takeoff_angle=15,
            takeoff_angle_errors={'uncertainty': 10.2},
            time_residual=.02,
            horizontal_slowness_residual=12.2,
            backazimuth_residual=12.2,
            time_weight=.23,
            horizontal_slowness_weight=12,
            backazimuth_weight=12,
            earth_model_id=ev.ResourceIdentifier(),
            commens=[ev.Comment(x) for x in 'Nothing'],
        )

    def _get_composite_times():
        return ev.CompositeTime(
            year=2016,
            year_errors={'uncertainty': 0},
            month=5,
            month_errors={'uncertainty': 0},
            day=4,
            day_errors={'uncertainty': 0},
            hour=0,
            hour_errors={'uncertainty': 0},
            minute=0,
            minute_errors={'uncertainty': 0},
            second=0,
            second_errors={'uncertainty': .01}
        )

    def _get_origin_quality():
        return ev.OriginQuality(
            associate_phase_count=1,
            used_phase_count=1,
            associated_station_count=1,
            used_station_count=1,
            depth_phase_count=1,
            standard_error=.02,
            azimuthal_gap=.12,
            ground_truth_level='GT0',
        )

    def _get_origin_uncertainty():
        return ev.OriginUncertainty(
            horizontal_uncertainty=1.2,
            min_horizontal_uncertainty=.12,
            max_horizontal_uncertainty=2.2,
            confidence_ellipsoid=_get_confidence_ellipsoid(),
            preferred_description="uncertainty ellipse",
        )

    def _get_confidence_ellipsoid():
        return ev.ConfidenceEllipsoid(
            semi_major_axis_length=12,
            semi_minor_axis_length=12,
            major_axis_plunge=12,
            major_axis_rotation=12,
        )

    def _create_magnitudes():
        return ev.Magnitude(
            resource_id=ev.ResourceIdentifier(),
            mag=5.5,
            mag_errors={'uncertainty': .01},
            magnitude_type='Mw',
            origin_id=state['origin_id'],
            station_count=1,
            station_magnitude_contributions=[_get_station_mag_contrib()],
        )

    def _get_station_mag():
        station_mag = ev.StationMagnitude(
            mag=2.24,
        )
        state['station_mag_id'] = station_mag.resource_id
        return station_mag

    def _get_station_mag_contrib():
        return ev.StationMagnitudeContribution(
            station_magnitude_id=state['station_mag_id'],
        )

    def _get_event_description():
        return ev.EventDescription(
            text='some text about the EQ',
            type='earthquake name',
        )

    def _get_amplitudes():
        return ev.Amplitude(
            generic_amplitude=.0012,
            type='A',
            unit='m',
            period=1,
            time_window=_get_timewindow(),
            pick_id=state['pick_id'],
            scalling_time=state['time'],
            mangitude_hint='ML',
            scaling_time_errors=ev.QuantityError(uncertainty=42.0),
        )

    def _get_timewindow():
        return ev.TimeWindow(
            begin=1.2,
            end=2.2,
            reference=UTCDateTime('2016-05-04T12:00:00'),
        )

    def _get_focal_mechanisms():
        return ev.FocalMechanism(
            nodal_planes=_get_nodal_planes(),
            principal_axis=_get_principal_axis(),
            azimuthal_gap=12,
            station_polarity_count=12,
            misfit=.12,
            station_distribution_ratio=.12,
            moment_tensor=_get_moment_tensor(),
        )

    def _get_nodal_planes():
        return ev.NodalPlanes(
            nodal_plane_1=ev.NodalPlane(strike=12, dip=2, rake=12),
            nodal_plane_2=ev.NodalPlane(strike=12, dip=2, rake=12),
            preferred_plane=2,
        )

    def _get_principal_axis():
        return ev.PrincipalAxes(
            t_axis=15,
            p_axis=15,
            n_axis=15,
        )

    def _get_moment_tensor():
        return ev.MomentTensor(
            scalar_moment=12213,
            tensor=_get_tensor(),
            variance=12.23,
            variance_reduction=98,
            double_couple=.22,
            clvd=.55,
            iso=.33,
            source_time_function=_get_source_time_function(),
            data_used=[_get_data_used()],
            method_id=ev.ResourceIdentifier(),
            inversion_type='general',
        )

    def _get_tensor():
        return ev.Tensor(
            m_rr=12,
            m_rr_errors={'uncertainty': .01},
            m_tt=12,
            m_pp=12,
            m_rt=12,
            m_rp=12,
            m_tp=12,
        )

    def _get_source_time_function():
        return ev.SourceTimeFunction(
            type='triangle',
            duration=.12,
            rise_time=.33,
            decay_time=.23,
        )

    def _get_data_used():
        return ev.DataUsed(
            wave_type='body waves',
            station_count=12,
            component_count=12,
            shortest_period=1,
            longest_period=20,
        )

    events = [_create_event()]
    return Catalog(events=events)


def streams_almost_equal(st1, st2, default_stats=True, rtol=1e-05, atol=1e-08,
                         equal_nan=True):
    """
    Return True if two streams are almost equal.

    :param st1: The first :class:`~obspy.core.stream.Stream` object.
    :param st2: The second :class:`~obspy.core.stream.Stream` object.
    :param default_stats:
        If True only compare the default stats on the traces, such as seed
        identification codes, start/end times, sampling_rates, etc. If
        False also compare extra stats attributes such as processing and
        format specific information.
    :param rtol: The relative tolerance parameter passed to
        :func:`~numpy.allclose` for comparing time series.
    :param atol: The absolute tolerance parameter passed to
        :func:`~numpy.allclose` for comparing time series.
    :param equal_nan:
        If ``True`` NaNs are evaluated equal when comparing the time
        series.
    :return: bool

    .. rubric:: Example

    1) Changes to the non-default parameters of the
        :class:`~obspy.core.trace.Stats` objects of the stream's contained
        :class:`~obspy.core.trace.Trace` objects will cause the streams to
        be considered unequal, but they will be considered almost equal.

        >>> from obspy import read
        >>> st1 = read()
        >>> st2 = read()
        >>> # The traces should, of course, be equal.
        >>> assert st1 == st2
        >>> # Perform detrending on st1 twice so processing stats differ.
        >>> st1 = st1.detrend('linear')
        >>> st1 = st1.detrend('linear')
        >>> st2 = st2.detrend('linear')
        >>> # The traces are no longer equal, but are almost equal.
        >>> assert st1 != st2
        >>> assert streams_almost_equal(st1, st2)


    2) Slight differences in each trace's data will cause the streams
        to be considered unequal, but they will be almost equal if the
        differences don't exceed the limits set by the ``rtol`` and
        ``atol`` parameters.

        >>> from obspy import read
        >>> st1 = read()
        >>> st2 = read()
        >>> # Perturb the trace data in st2 slightly.
        >>> for tr in st2:
        ...     tr.data *= (1 + 1e-6)
        >>> # The streams are no longer equal.
        >>> assert st1 != st2
        >>> # But they are almost equal.
        >>> assert streams_almost_equal(st1, st2)
        >>> # Unless, of course, there is a large change.
        >>> st1[0].data *= 10
        >>> assert not streams_almost_equal(st1, st2)
    """
    from obspy.core.stream import Stream
    # Return False if both objects are not streams or not the same length.
    are_streams = isinstance(st1, Stream) and isinstance(st2, Stream)
    if not are_streams or not len(st1) == len(st2):
        return False
    # Kwargs to pass trace_almost_equal.
    tr_kwargs = dict(default_stats=default_stats, rtol=rtol, atol=atol,
                     equal_nan=equal_nan)
    # Ensure the streams are sorted (as done with the __equal__ method)
    st1_sorted = st1.select()
    st1_sorted.sort()
    st2_sorted = st2.select()
    st2_sorted.sort()
    # Iterate over sorted trace pairs and determine if they are almost equal.
    for tr1, tr2 in zip(st1_sorted, st2_sorted):
        if not traces_almost_equal(tr1, tr2, **tr_kwargs):
            return False  # If any are not almost equal return None.
    return True


def traces_almost_equal(tr1, tr2, default_stats=True, rtol=1e-05, atol=1e-08,
                        equal_nan=True):
    """
    Return True if the two traces are almost equal.

    :param tr1: The first :class:`~obspy.core.trace.Trace` object.
    :param tr2: The second :class:`~obspy.core.trace.Trace` object.
    :param default_stats:
        If True only compare the default stats on the traces, such as seed
        identification codes, start/end times, sampling_rates, etc. If
        False also compare extra stats attributes such as processing and
        format specific information.
    :param rtol: The relative tolerance parameter passed to
        :func:`~numpy.allclose` for comparing time series.
    :param atol: The absolute tolerance parameter passed to
        :func:`~numpy.allclose` for comparing time series.
    :param equal_nan:
        If ``True`` NaNs are evaluated equal when comparing the time
        series.
    :return: bool
    """
    from obspy.core.trace import Trace
    # If other isnt  a trace, or data is not the same len return False.
    if not isinstance(tr2, Trace) or len(tr1.data) != len(tr2.data):
        return False
    # First compare the array values
    try:  # Use equal_nan if available
        all_close = np.allclose(tr1.data, tr2.data, rtol=rtol,
                                atol=atol, equal_nan=equal_nan)
    except TypeError:
        # This happens on very old versions of numpy. Essentially
        # we just need to handle NaN detection on our own, if equal_nan.
        is_close = np.isclose(tr1.data, tr2.data, rtol=rtol, atol=atol)
        if equal_nan:
            isnan = np.isnan(tr1.data) & np.isnan(tr2.data)
        else:
            isnan = np.zeros(tr1.data.shape).astype(bool)
        all_close = np.all(isnan | is_close)
    # Then compare the stats objects
    stats1 = _make_stats_dict(tr1, default_stats)
    stats2 = _make_stats_dict(tr2, default_stats)
    return all_close and stats1 == stats2


def _make_stats_dict(tr, default_stats):
    """
    Return a dict of stats from trace optionally including processing.
    """
    from obspy.core.trace import Stats
    if not default_stats:
        return dict(tr.stats)
    return {i: tr.stats[i] for i in Stats.defaults}


if __name__ == '__main__':
    doctest.testmod(exclude_empty=True)



import obspy
from obspy.signal.trigger import delayed_sta_lta, plot_trigger


trace = obspy.read("https://examples.obspy.org/ev0_6.a01.gse2")[0]
df = trace.stats.sampling_rate

cft = delayed_sta_lta(trace.data, int(5 * df), int(10 * df))
plot_trigger(trace, cft, 5, 10)



# -*- coding: utf-8 -*-
import io
import re

import pytest
import warnings
from obspy import read_events
from obspy.io.iaspei.core import _read_ims10_bulletin, _is_ims10_bulletin
from obspy.core.event import ResourceIdentifier
from obspy.core.util.obspy_types import ObsPyReadingError


class TestIASPEI():
    """
    Test suite for obspy.io.iaspei.core
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, testdata):
        # XXX the converted QuakeML file is not complete.. many picks and
        # station magnitudes were removed from it to save space.
        self.path_to_ims = testdata['19670130012028.isf']
        self.path_to_quakeml = testdata['19670130012028.xml']
        self.path_to_ims_2 = testdata["ipe202409sel_ims.txt"]
        self.path_to_quakeml_2 = testdata["ipe202409sel_ims.xml"]

    def prepare_comparison(self, catalog):
        # some more fixes for the comparison, these are due to buggy QuakeML
        # reader behavior and should be fixed in io.quakeml eventually
        for event in catalog:
            for pick in event.picks:
                # QuakeML reader seems to set `network_code=""` if it's not in
                # the xml file.. account for this strange behavior for this
                # test
                pick.waveform_id.network_code = None
                # QuakeML reader seems to add empty QuantityError for
                # pick.horizontal_slowness_errors
                for key in ['horizontal_slowness_errors', 'time_errors',
                            'backazimuth_errors']:
                    setattr(pick, key, None)
            for origin in event.origins:
                if origin.origin_uncertainty is not None:
                    # QuakeML reader seems to add empty ConfidenceEllipsoid
                    origin.origin_uncertainty.confidence_ellipsoid = None
                # QuakeML reader seems to add empty QuantityError for
                # pick.horizontal_slowness_errors
                for key in ['longitude_errors', 'latitude_errors',
                            # 'depth_errors', 'time_errors',
                            ]:
                    setattr(origin, key, None)
            for station_magnitude in event.station_magnitudes:
                setattr(station_magnitude.waveform_id, 'network_code', None)
                # QuakeML reader seems to set origin_id to
                # `ResourceIdentifier(id="None")`
                if station_magnitude.origin_id == ResourceIdentifier(
                        id="None"):
                    setattr(station_magnitude, 'origin_id', None)
                # QuakeML reader seems to add empty QuantityError for
                # pick.horizontal_slowness_errors
                for key in ['mag_errors']:
                    setattr(station_magnitude, key, None)
            for magnitude in event.magnitudes:
                # QuakeML reader seems to add empty QuantityError for
                # pick.horizontal_slowness_errors
                for key in ['mag_errors']:
                    setattr(magnitude, key, None)

    def _assert_catalog(self, got):
        got_id_prefix = str(got.resource_id).encode('UTF-8')
        # first of all, replace the random hash in our test file with the
        # prefix that was used during reading the ISF file
        with open(self.path_to_quakeml, 'rb') as fh:
            data = fh.read()
        match = re.search(b'publicID="(smi:local/[a-z0-9-]*)"', data)
        expected_id_prefix = match.group(1)
        data, num_subs = re.subn(expected_id_prefix, got_id_prefix, data)
        # resource id replacements should be done in the QuakeML file we
        # compare against
        assert num_subs == 54
        bio = io.BytesIO(data)
        expected = read_events(bio, format="QUAKEML")
        # now first check if we got the expected number of picks and station
        # magnitudes, because the quakeml file has been stripped of picks and
        # station magnitudes to save space and time
        assert len(got[0].picks) == 255
        assert len(got[0].station_magnitudes) == 15
        # ok now crop the got catalog accordingly, afterwards it should compare
        # equal to our comparison catalog
        got[0].picks = got[0].picks[:4]
        got[0].station_magnitudes = got[0].station_magnitudes[:7]
        for origin in got[0].origins:
            if origin.arrivals:
                origin.arrivals = origin.arrivals[:4]
        self.prepare_comparison(expected)
        # now finally these catalogs should compare equal
        assert got == expected

    def _assert_catalog2(self, got):
        # Unify the random hash prefix of all resource_ids on the side
        # of the ISF test file and on the QuakeML model file side.
        # Resource id replacements should be done in the QuakeML file.
        got_id_prefix = str(got.resource_id).encode('UTF-8')
        with open(self.path_to_quakeml_2, 'rb') as fh:
            data = fh.read()
        match = re.search(b'publicID="(smi:local/[a-z0-9-]*)"', data)
        expected_id_prefix = match.group(1)
        data, num_subs = re.subn(expected_id_prefix, got_id_prefix, data)
        bio = io.BytesIO(data)
        expected = read_events(bio, format="QUAKEML")
        self.prepare_comparison(expected)
        # got.write("DEBUGgot.xml", format="QUAKEML")
        # expected.write("DEBUGexpected.xml", format="QUAKEML")
        # take the entire uncropped got catalog and compare against expected
        assert got == expected

    def test_reading(self):
        """
        Test reading IMS10 bulletin format
        """
        cat = _read_ims10_bulletin(self.path_to_ims, _no_uuid_hashes=True)
        assert len(cat) == 1
        self._assert_catalog(cat)

    def test_reading_via_file(self):
        """
        Test reading IMS10 bulletin format from open files.
        """
        with io.open(self.path_to_ims, "rb") as fh:
            cat = _read_ims10_bulletin(fh, _no_uuid_hashes=True)
        assert len(cat) == 1
        self._assert_catalog(cat)

        with io.open(self.path_to_ims, "rt", encoding="UTF-8") as fh:
            cat = _read_ims10_bulletin(fh, _no_uuid_hashes=True)
        assert len(cat) == 1
        self._assert_catalog(cat)

    def test_reading_via_bytes_io(self):
        """
        Test reading IMS10 bulletin format from bytes io object.
        """
        with io.open(self.path_to_ims, "rb") as fh:
            with io.BytesIO(fh.read()) as buf:
                buf.seek(0, 0)
                cat = _read_ims10_bulletin(buf, _no_uuid_hashes=True)
        assert len(cat) == 1
        self._assert_catalog(cat)

    def test_reading_via_plugin(self):
        """
        Test reading IMS10 bulletin format
        """
        cat = read_events(self.path_to_ims, format='IMS10BULLETIN',
                          _no_uuid_hashes=True)
        assert len(cat) == 1
        self._assert_catalog(cat)

    def test_is_ims10_bulltin(self):
        """
        Test checking if file is IMS10 bulletin format
        """
        assert _is_ims10_bulletin(self.path_to_ims)
        assert not _is_ims10_bulletin(self.path_to_quakeml)

    def test_is_ims10_bulltin_open_file(self):
        with open(self.path_to_ims, "rb") as fh:
            assert _is_ims10_bulletin(fh)

        with open(self.path_to_ims, "rt", encoding="utf-8") as fh:
            assert _is_ims10_bulletin(fh)

        with open(self.path_to_quakeml, "rb") as fh:
            assert not _is_ims10_bulletin(fh)

        with open(self.path_to_quakeml, "rt", encoding="utf-8") as fh:
            assert not _is_ims10_bulletin(fh)

    def test_is_ims10_bulltin_from_bytes_io(self):
        with open(self.path_to_ims, "rb") as fh:
            with io.BytesIO(fh.read()) as buf:
                buf.seek(0, 0)
                assert _is_ims10_bulletin(buf)

        with open(self.path_to_quakeml, "rb") as fh:
            with io.BytesIO(fh.read()) as buf:
                buf.seek(0, 0)
                assert not _is_ims10_bulletin(buf)

    def test_reading_2(self):
        """
        Test reading IMS10 bulletin format
        """
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # The last of the three events
            # has a phase block with incorrect #OriginID tag
            cat = _read_ims10_bulletin(self.path_to_ims_2,
                                       _no_uuid_hashes=True,
                                       skip_orphan=False,
                                       origin_specific_to_comments=True)
            assert len(w) == 3
            assert issubclass(w[-3].category, UserWarning)
            s_msg = "Phase block cannot be fully processed"
            assert s_msg in str(w[-3].message)
            assert issubclass(w[-2].category, UserWarning)
            s_msg = "This pick would have a time more than 6 hours after"
            assert s_msg in str(w[-2].message)
            assert issubclass(w[-1].category, UserWarning)
            s_msg = "Could not determine absolute time of pick"
            assert s_msg in str(w[-1].message)
        assert len(cat) == 3
        self._assert_catalog2(cat)

    def test_reading_file_faulty(self):
        """
        Test reading IMS10 bulletin from a file of non-compliant format
        """
        fbuf = io.BytesIO()
        with open(self.path_to_ims_2, 'rb') as fi:
            for line in fi:
                if line.startswith(b"DATA_TYPE BULLETIN IMS1.0"):
                    # Corrupt the DATA_TYPE line
                    line = b"DATA_TYPE BULLETIN IMS1.0:long\n"
                fbuf.write(line)
        # print(fbuf.getvalue().decode('utf-8'))
        fbuf.seek(0, 0)
        assert not _is_ims10_bulletin(fbuf)
        fbuf.seek(0, 0)
        with pytest.raises(ObsPyReadingError) as excinfo:
            _read_ims10_bulletin(fbuf)
            s_msg = "No data section of valid DATA_TYPE found"
            print(excinfo.value)
            assert s_msg in str(excinfo.value)



from __future__ import print_function

import numpy as np
from obspy import UTCDateTime, read, Trace, Stream


weather = """
00.0000 0.0 ??? 4.7 97.7 1015.0 0.0 010308 000000
00.0002 0.0 ??? 4.7 97.7 1015.0 0.0 010308 000001
00.0005 0.0 ??? 4.7 97.7 1015.0 0.0 010308 000002
00.0008 0.0 ??? 4.7 97.7 1015.4 0.0 010308 000003
00.0011 0.0 ??? 4.7 97.7 1015.0 0.0 010308 000004
00.0013 0.0 ??? 4.7 97.7 1015.0 0.0 010308 000005
00.0016 0.0 ??? 4.7 97.7 1015.0 0.0 010308 000006
00.0019 0.0 ??? 4.7 97.7 1015.0 0.0 010308 000007
"""

# Convert to NumPy character array
data = np.fromstring(weather, dtype='|S1')

# Fill header attributes
stats = {'network': 'BW', 'station': 'RJOB', 'location': '',
         'channel': 'WLZ', 'npts': len(data), 'sampling_rate': 0.1,
         'mseed': {'dataquality': 'D'}}
# set current time
stats['starttime'] = UTCDateTime()
st = Stream([Trace(data=data, header=stats)])
# write as ASCII file (encoding=0)
st.write("weather.mseed", format='MSEED', encoding=0, reclen=256)

# Show that it worked, convert NumPy character array back to string
st1 = read("weather.mseed")
print(st1[0].data.tobytes())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A command-line program that runs all ObsPy tests.

All tests in ObsPy are located in the tests directory of each specific
module.

A few important command line arguments:
    --help : Print help message an exit.
    --no-report, --report : Do not ask and automatically (not) submit report
            to ObsPys' report server.
    --cov obspy: Calculate and display test coverage.
    --network : Only run network tests (requires an internet connection).
    --all : Run both network and non-network tests.
    --keep-images : Store all images generated by tests in obspy's directory.

Also note, the test runner uses pytest under the hood so any pytest command
line argument is also accepted.

.. rubric:: Examples

(1) Run all local tests (ignoring tests requiring a network connection) on
    command line::
        $ obspy-runtests
(2) Run all tests on command line (including network tests)::
        $ obspy-runtests --all
(3) Run tests of module :mod:`obspy.io.mseed`::
        $ obspy-runtests io/mseed
(4) Run tests of multiple modules, e.g. :mod:`obspy.io.wav` and
    :mod:`obspy.io.sac`::
        $ obspy-runtests io/wav obspy/io/sac
(5) Run a specific test case::
        $ obspy-runtests core/tests/test_stats.py::TestStats::test_init
(6) Run tests and print a coverage report to screen including missed lines:
        $ obspy-runtests --cov obspy --cov-report term-missing
(7) Save the image outputs of the testsuite, called 'obspy_image_tests':
        $ obspy-runtests --keep-images
(8) Run the test suite, drop into a pdb debugging session for each failure:
        $ obspy-runtests --pdb

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import contextlib
import os
import sys
from pathlib import Path


# URL to upload json report
REPORT_URL = "tests.obspy.org"


@contextlib.contextmanager
def change_directory(path):
    """
    A context manager to change directory to target path.

    :param path: The path to change to.
    :type path: A string or pathlib Path.
    """
    origin = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)


def main():
    """
    Entry point for setup.py.
    Wrapper for a profiler if requested otherwise just call run() directly.
    If profiling is enabled we disable interactivity as it would wait for user
    input and influence the statistics. However the -r option still works.
    """
    if '-h' in sys.argv or '--help' in sys.argv:
        print(__doc__)
        sys.exit(0)
    try:
        import pytest
        from pytest_jsonreport.plugin import JSONReport
    except ImportError:
        msg = ("\nObsPy's test suite uses pytest and pytest-json-report. "
               "Please install these packages before using obspy-runtests. "
               "Example with pip:\n"
               "\t$ pip install pytest pytest-json-report\n\n"
               "Additional optional test requirements can be used and "
               "installed with:\n"
               "\t$ pip install obspy[tests]\n")
        sys.exit(msg)
    # hack to get rid of internal pytest warning, see
    # https://github.com/pytest-dev/pytest-cov/issues/148
    import pytest_jsonreport
    pytest_jsonreport.__doc__ = 'PYTEST_DONT_REWRITE'
    # Cd into ObsPy's directory and run tests.
    with change_directory(Path(__file__).parent.parent):
        plugin = JSONReport()
        status = pytest.main(plugins=[plugin])
    # Upload test report if tests were successfully run.
    # See https://docs.pytest.org/en/latest/reference/exit-codes.html
    if int(status) < 2:
        report = (True if '--report' in sys.argv else
                  False if '--no-report' in sys.argv else None)
        upload_json_report(report=report, data=plugin.report)
    sys.exit(status)


def upload_json_report(report=None, data=None):
    """Upload the json report to ObsPy test server."""
    import requests
    if report is None:
        msg = f"Do you want to report this to {REPORT_URL} ? [n]: "
        answer = input(msg).lower()
        report = 'y' in answer
    if report:
        # only include unique warnings.
        data.setdefault('warnings', [])
        data['warnings'] = [
            dict(t) for t in {tuple(d.items()) for d in data['warnings']}
        ]
        response = requests.post(f"https://{REPORT_URL}/post/v2/", json=data)
        # get the response
        if response.status_code == 200:
            report_url = response.json().get('url', REPORT_URL)
            print('Your test results have been reported and are available at: '
                  '{}\nThank you!'.format(report_url))
        # handle errors
        else:
            print(f"Error: Could not sent a test report to {REPORT_URL}.")
            print(response.reason)
            print(response.content)


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib.pyplot as plt

import obspy

# Read the seismogram
st = obspy.read("https://examples.obspy.org/RJOB_061005_072159.ehz.new")

# There is only one trace in the Stream object, let's work on that trace...
tr = st[0]

# Decimate the 200 Hz data by a factor of 4 to 50 Hz. Note that this
# automatically includes a lowpass filtering with corner frequency 20 Hz.
# We work on a copy of the original data just to demonstrate the effects of
# downsampling.
tr_new = tr.copy()
tr_new.decimate(factor=4, strict_length=False)

# For comparison also only filter the original data (same filter options as in
# automatically applied filtering during downsampling, corner frequency
# 0.4 * new sampling rate)
tr_filt = tr.copy()
tr_filt.filter('lowpass', freq=0.4 * tr.stats.sampling_rate / 4.0)

# Now let's plot the raw and filtered data...
t = np.arange(0, tr.stats.npts / tr.stats.sampling_rate, tr.stats.delta)
t_new = np.arange(0, tr_new.stats.npts / tr_new.stats.sampling_rate,
                  tr_new.stats.delta)

plt.plot(t, tr.data, 'k', label='Raw', alpha=0.3)
plt.plot(t, tr_filt.data, 'b', label='Lowpassed', alpha=0.7)
plt.plot(t_new, tr_new.data, 'r', label='Lowpassed/Downsampled', alpha=0.7)
plt.xlabel('Time [s]')
plt.xlim(82, 83.5)
plt.suptitle(tr.stats.starttime)
plt.legend()
plt.show()



# -*- coding: utf-8 -*-
from .blockette import Blockette
from ..fields import FixedString, Float, Integer, VariableString


class Blockette050(Blockette):
    """
    Blockette 050: Station Identifier Blockette.

    Sample:
    0500097ANMO +34.946200-106.456700+1740.00006001Albuquerque, NewMexico, USA~
    0013210101989,241~~NIU
    """

    id = 50
    name = "Station Identifier"
    fields = [
        FixedString(3, "Station call letters", 5, 'UN'),
        Float(4, "Latitude", 10, mask='%+02.6f'),
        Float(5, "Longitude", 11, mask='%+03.6f'),
        Float(6, "Elevation", 7, mask='%+04.1f'),
        Integer(7, "Number of channels", 4),
        Integer(8, "Number of station comments", 3),
        VariableString(9, "Site name", 1, 60, 'UNLPS'),
        Integer(10, "Network identifier code", 3, xpath=33),
        Integer(11, "word order 32bit", 4),
        Integer(12, "word order 16bit", 2),
        VariableString(13, "Start effective date", 1, 22, 'T'),
        VariableString(14, "End effective date", 0, 22, 'T', optional=True,
                       xseed_version='1.0'),
        VariableString(14, "End effective date", 0, 22, 'T',
                       xseed_version='1.1'),
        FixedString(15, "Update flag", 1),
        FixedString(16, "Network Code", 2, 'ULN', version=2.3)
    ]
